"""
Module for AWSConfig class.
"""

from configparser import ConfigParser, NoOptionError, NoSectionError
from os import path as osp


class AWSConfig:
    """
    Class AWSConfig parses AWS CLI config file, ``~/.aws/config`` by default,
    and provides a convenient interfaces to certain configuration options.

    :param aws_home: Path to a directory with AWS configs. By default, ``~/.aws/``.
    :type aws_home: str
    """

    def __init__(self, aws_home=None):
        self._aws_home = aws_home
        self._config_parser = None

    @property
    def aws_home(self):
        """Path to AWS config directory."""
        if self._aws_home is None:
            self._aws_home = osp.join(osp.expanduser("~"), ".aws")
        return self._aws_home

    @property
    def config_path(self):
        """Path to AWS config file."""
        return osp.join(self.aws_home, "config")

    @property
    def config_parser(self) -> ConfigParser:
        """ConfigParser object that represents ``~/.aws/config``."""
        if self._config_parser is None:
            self._config_parser = ConfigParser()
            self._config_parser.read(self.config_path)

        return self._config_parser

    @property
    def profiles(self) -> list:
        """List of configured AWS profiles."""
        return [
            "default" if section == "default" else section.split(" ")[1]
            for section in self.config_parser.sections()
            if (section.startswith("profile") or section == "default")
        ]

    @property
    def regions(self) -> list:
        """Return a list of all AWS regions"""
        return [
            "af-south-1",
            "ap-east-1",
            "ap-northeast-1",
            "ap-northeast-2",
            "ap-northeast-3",
            "ap-south-1",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-southeast-3",
            "ca-central-1",
            "eu-central-1",
            "eu-north-1",
            "eu-south-1",
            "eu-west-1",
            "eu-west-2",
            "eu-west-3",
            "me-south-1",
            "sa-east-1",
            "us-east-1",
            "us-east-2",
            "us-west-1",
            "us-west-2",
        ]

    def get_account_id(self, profile_name):
        """Read account id for given profile."""
        return self.config_parser.get(self._get_section(profile_name), "sso_account_id")

    def get_region(self, profile_name):
        """Read AWS region for given profile."""
        try:
            return self.config_parser.get(self._get_section(profile_name), "region")
        except NoSectionError:
            return self.config_parser.get("default", "region") if "default" in self.config_parser.sections() else None

        except NoOptionError:
            return self.config_parser.get("default", "region") if "default" in self.config_parser.sections() else None

    def get_role(self, profile_name):
        """Read AWS IAM role for given profile."""
        return self.config_parser.get(self._get_section(profile_name), "sso_role_name")

    def get_start_url(self, profile_name):
        """Read SSO URL for given profile."""
        sso_session = self.config_parser.get(self._get_section(profile_name), "sso_session")
        return self.config_parser.get(f"sso-session {sso_session}", "sso_start_url")

    @staticmethod
    def _get_section(profile_name):
        return "default" if profile_name in ["default", None] else f"profile {profile_name}"



import base64
import json

from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


def test_metadata():
    backend = TFS3Backend("foo_backet", "path/to/tf.state")
    status = TFStatus(backend, True, RunResult(1, 2, 3), RunOutput("no stdout", "no stderr"))
    assert isinstance(status.metadata, str)
    assert json.loads(base64.b64decode(status.metadata))[backend.id]["add"] == 1
    assert json.loads(base64.b64decode(status.metadata))[backend.id]["change"] == 2
    assert json.loads(base64.b64decode(status.metadata))[backend.id]["destroy"] == 3



"""
.. topic:: ``ih-elastic snapshots restore``

    A ``ih-elastic snapshots restore`` subcommand.

    See ``ih-elastic snapshots restore --help`` for more details.
"""

from logging import getLogger

import click
from elasticsearch.client import SnapshotClient

LOG = getLogger(__name__)


@click.command(name="restore")
@click.argument("repository-name")
@click.argument("snapshot-name")
@click.pass_context
def cmd_restore(ctx, **kwargs):
    """
    Restores a snapshot in a repository.
    """
    client = SnapshotClient(ctx.obj["es"])
    client.restore(repository=kwargs["repository_name"], snapshot=kwargs["snapshot_name"])



"""
.. topic:: ``ih-aws ecs``

    A ``ih-aws ecs`` subcommand.

    See ``ih-aws ecs --help`` for more details.
"""

import sys
import time
from logging import getLogger

import click
from botocore.exceptions import ClientError

from infrahouse_toolkit.aws import get_aws_client

LOG = getLogger(__name__)


@click.command(name="wait-services-stable")
@click.option("--cluster", help="ECS cluster name that runs requested services.", required=True)
@click.option(
    "--service",
    help="ECS service name that we wait to become stable. Multiple services can be specified.",
    required=True,
    multiple=True,
)
@click.option(
    "--wait-timeout",
    help="Time in seconds to wait until all services become stable.",
    default=20 * 60,
    show_default=True,
)
@click.pass_context
def cmd_wait_services_stable(ctx, **kwargs):
    """
    Wait up to --timeout seconds until all specified services in an ECS cluster
    become stable.

    The service is considered stable when there is only one deployment for it
    and number of running tasks is equal to desired number tasks.
    """
    aws_config = ctx.obj["aws_config"]
    aws_profile = ctx.obj["aws_profile"]
    cluster = kwargs["cluster"]
    services = kwargs["service"]
    wait_timeout = kwargs["wait_timeout"]
    try:
        aws_session = ctx.obj["aws_session"]
        response = get_aws_client(
            "sts", aws_profile, aws_config.get_region(aws_profile), session=aws_session
        ).get_caller_identity()
        LOG.debug("Connected to AWS as %s", response["Arn"])
        ecs_client = get_aws_client("ecs", aws_profile, aws_config.get_region(aws_profile), session=aws_session)
        now = time.time()
        _wait_for_services(ecs_client, cluster, services, start_time=now, end_time=now + wait_timeout)

    except ClientError as err:
        LOG.exception(err)
        LOG.error("Try to run ih-aws with --aws-profile option.")
        LOG.error("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)


def _wait_for_services(ecs_client, cluster, services, start_time, end_time):
    sleep_time = 15
    while time.time() < end_time:
        response = ecs_client.describe_services(cluster=cluster, services=list(services))

        all_services_stable = True
        for service in response["services"]:
            n_deployments = len(service["deployments"])
            LOG.info("Service: %s, deployments = %d", service["serviceName"], n_deployments)

            for deployment in service["deployments"]:
                LOG.info(
                    "Service: %s, deployment = %s (%s), running = %d, desired = %s",
                    service["serviceName"],
                    deployment["id"],
                    deployment["status"],
                    deployment["runningCount"],
                    deployment["desiredCount"],
                )
                if deployment["runningCount"] != deployment["desiredCount"]:
                    all_services_stable = False

            if n_deployments != 1:
                all_services_stable = False

        if all_services_stable:
            LOG.info("Services %s are stable", ",".join(services))
            sys.exit(0)
        else:
            LOG.info("Services are not stable yet. Waiting %s seconds", sleep_time)
            time.sleep(sleep_time)

    LOG.error("Services %s didn't become stable after %s seconds", ",".join(services), end_time - start_time)
    sys.exit(1)



"""
.. topic:: ``ih-github scan``

    A ``ih-github scan`` subcommand.

    See ``ih-github scan --help`` for more details.
"""

import json
import logging
import sys
from datetime import datetime, timedelta
from os import path as osp
from subprocess import PIPE, Popen

import click

from infrahouse_toolkit.cli.utils import check_dependencies
from infrahouse_toolkit.terraform.githubpr import GitHubPR

LOG = logging.getLogger()
DEPENDENCIES = ["osv-scanner"]


@click.command(
    name="scan",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@click.option("--github-token", help="Personal access token for GitHub.", envvar="GITHUB_TOKEN")
@click.option(
    "--osv-config", help="Path to osv-scanner configuration file.", default="osv-scanner.toml", show_default=True
)
@click.option("--repo", help="GitHub repository name in a long format e.g. infrahouse/infrahouse-toolkit.")
@click.option("--pull-request", help="GitHub pull request number.", type=click.INT)
@click.option(
    "--sla-critical",
    help="SLA in days on when critical vulnerabilities should be fixed.",
    default=15,
    show_default=True,
)
@click.option(
    "--sla-high", help="SLA in days on when high vulnerabilities should be fixed.", default=30, show_default=True
)
@click.option(
    "--sla-medium", help="SLA in days on when medium vulnerabilities should be fixed.", default=45, show_default=True
)
@click.option(
    "--sla-low", help="SLA in days on when medium vulnerabilities should be fixed.", default=90, show_default=True
)
@click.pass_context
def cmd_scan(ctx, *args, **kwargs):
    """
    Scan the current directory for dependency vulnerabilities.
    If found, publish a report as a pull request comment.
    """
    LOG.debug("args = %s", args)
    LOG.debug("kwargs = %s", kwargs)
    cmd_args = ctx.args
    LOG.debug("cmd_args = %s", cmd_args)
    check_dependencies(DEPENDENCIES)
    osv_config = kwargs["osv_config"]
    cmd = ["osv-scanner", "scan", "--format", "markdown", "--recursive", "--verbosity", "warn"]
    if osp.exists(osv_config):
        cmd.extend(["--config", osv_config])
    cmd.extend(cmd_args)
    cmd.extend(["./"])
    sla_map = {
        "CRITICAL": kwargs["sla_critical"],
        "HIGH": kwargs["sla_high"],
        "MODERATE": kwargs["sla_medium"],
        "MEDIUM": kwargs["sla_medium"],
        "LOW": kwargs["sla_low"],
    }
    comment = None
    try:
        with Popen(cmd, stderr=PIPE, stdout=PIPE) as proc:
            LOG.info("Launched command: %s", " ".join(cmd))
            vuln_table, cerr = proc.communicate()
            return_code = proc.returncode

            if cerr:
                LOG.error(cerr.decode())

            if return_code == 1:
                comment = "# Vulnerabilities report\n"
                comment += vuln_table.decode()
                for vuln in _get_vulnerability_details(osv_config, cmd_args):
                    future_date = datetime.now() + timedelta(days=sla_map.get(vuln["severity"], 30))
                    comment += f"""
## {vuln["name"]}=={vuln["version"]}
The package `{vuln["name"]}` (version {vuln["version"]}) has a {vuln["severity"]}-severity vulnerability.

If you're able to upgrade to a non-vulnerable version, we recommend doing so.
If upgrading isn't currently possible, consider adding this entry to your `{osv_config}` file
in the root directory of this repository.

```
[[IgnoredVulns]]
id = "{vuln["id"]}"
ignoreUntil = {future_date.strftime("%Y-%m-%d")} # Optional exception expiry date
reason = "Detailed explanation why the vulnerability is ignored and how it is planned to be fixed."
```
"""
            elif return_code == 128:
                LOG.warning("No package lock file found, will assume no vulnerabilities.")
                sys.exit(0)
            else:
                LOG.warning("osv-scanner exited with unknown exit code %d", return_code)

            sys.stdout.write(vuln_table.decode())

    finally:
        if comment:
            print(comment)
            if kwargs["repo"] and kwargs["pull_request"]:
                pull_request = GitHubPR(kwargs["repo"], kwargs["pull_request"], github_token=kwargs["github_token"])
                pull_request.publish_comment(comment)

    sys.exit(return_code)


def _get_vulnerability_details(config_file, extra_args):
    """
    returns a JSON
    [
        {
            name: "pymysql"
            version: "1.2.3"
            id: "GHSA-7wqh-767x-r66v"
            severity: "MODERATE"
        }
    ]
    """
    cmd = ["osv-scanner", "scan", "--format", "json", "--recursive", "--verbosity", "warn"]
    if osp.exists(config_file):
        cmd.extend(["--config", config_file])
    cmd.extend(extra_args)
    cmd.extend(["./"])
    return_value = []
    with Popen(cmd, stderr=PIPE, stdout=PIPE) as proc:
        LOG.debug("Launched command: %s", " ".join(cmd))
        cout, cerr = proc.communicate()
        if cerr:
            LOG.error(cerr)
        response = json.loads(cout)
        for result in response["results"]:
            for package_item in result["packages"]:
                for vuln in package_item["vulnerabilities"]:
                    return_value.append(
                        {
                            "name": package_item["package"]["name"],
                            "version": package_item["package"]["version"],
                            "id": vuln["id"],
                            "severity": vuln["database_specific"]["severity"],
                        }
                    )
        return return_value



"""Top level exceptions.

The exception hierarchy repeats the structure of the infrahouse_toolkit package.
Each module in the package has its own exceptions.py module.
The module exceptions are inherited from the upper module exceptions.

"""


class IHException(Exception):
    """Generic InfraHouse exception"""


class IHRetriableError(IHException):
    """Operation failed, but can be retried."""

    def __init__(self, returncode=None, cmd=None, output=None, stderr=None):
        self.returncode = returncode
        self.cmd = " ".join(cmd) if isinstance(cmd, list) else cmd
        self.output = output
        self.stderr = stderr

    def __str__(self):
        return f"Command '{self.cmd}' returned non-zero exit status {self.returncode}."



"""
Module for :py:class:`TFStatus`, Terraform plan run status class.
"""

import json
import logging
import re
from base64 import b64decode, b64encode
from collections import namedtuple
from difflib import unified_diff

from tabulate import tabulate

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.terraform.backends.tfbackend import TFBackend

RunResult = namedtuple("RunResult", "add change destroy")
RunOutput = namedtuple("RunOutput", "stdout stderr")

LOG = logging.getLogger()

RE_NO_COLOR = r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
"""


def decolor(text: str) -> str:
    """Remove ANSI escape sequences that color console output."""
    if text:
        ansi_escape = re.compile(RE_NO_COLOR, re.VERBOSE)
        return ansi_escape.sub("", text)

    return text


def strip_lines(src: str, pattern: str) -> str:
    """
    Remove lines starting with a string ``pattern``.

    :param src: Input text
    :type src: str
    :param pattern: A string. When a line in the input text starts with this string - skip it.
    :type pattern: str
    :return: Stripped text.
    :rtype: str
    """
    return (
        "\n".join([x for x in src.splitlines() if not x.startswith(pattern)]) + ("\n" if src[-1] == "\n" else "")
        if src
        else src
    )


class TFStatus:
    """
    :py:class:`TFStatus` represents a result of a ``terraform plan`` run.
    It includes outputs (both stdout and stderr) and a summary of changes -
    how many resources are going to be created/changed/destroyed.

    Credit for emojis https://emojicombos.com/
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    # Probably counts could be calculated from
    # affected_resources, but it's optional.
    def __init__(
        self,
        backend: TFBackend,
        success: bool,
        run_result: RunResult,
        run_output: RunOutput,
        affected_resources: RunResult = None,
    ):
        self.backend = backend
        self.success = success
        self.add = run_result.add
        self.change = run_result.change
        self.destroy = run_result.destroy
        self.stdout = run_output.stdout
        self.stderr = run_output.stderr
        self.affected_resources = affected_resources

    @property
    def comment(self):
        """Serialize the status as a comment text eligible to be posted on GitHub."""
        return (
            f"\n# State **`{self.backend.id}`**\n"
            + f"## Affected resources counts\n\n{self.summary_counts}\n"
            + (f"## Affected resources by action\n\n{self.summary_resources}\n" if self.affected_resources else "")
            + f"""<details>\n<summary>STDOUT</summary>\n\n```\n{self._short_stdout or "no output"}\n```\n</details>\n"""
            + f"""<details><summary><i>metadata</i></summary>\n\n```\n{self.metadata}\n```\n</details>"""
        )

    @property
    def metadata(self):
        """
        Produces a base64 encoded string with a dictionary that can be used
        to create the same instance of the class.
        """
        return b64encode(str(self).encode(DEFAULT_OPEN_ENCODING)).decode(DEFAULT_OPEN_ENCODING)

    @property
    def summary_counts(self):
        """
        Credit for tabulate:
        https://stackoverflow.com/questions/9535954/printing-lists-as-tabular-data

        :return: Formatted table.
        """
        rows = [
            [
                "✅" if self.success else "❌",
                self.add if self.success and self.add is not None else "❔",
                self.change if self.success and self.change is not None else "❔",
                self.destroy if self.success and self.destroy is not None else "❔",
            ]
        ]
        return tabulate(
            rows,
            headers=[
                "Success",
                f"{'🟢' if self.success and self.add is not None and self.add > 0 else ''} Add",
                f"{'🟡' if self.success and self.change is not None and self.change > 0 else ''} Change",
                f"{'🔴' if self.success and self.destroy is not None and self.destroy > 0 else ''} Destroy",
            ],
            colalign=("right",),
            tablefmt="pipe",
        )

    @property
    def summary_resources(self):
        """
        Produces a string with a table that lists all added/modified/deleted resources.
        """
        if all(
            (
                self.affected_resources,
                self.affected_resources.add is not None,
                self.affected_resources.change is not None,
                self.affected_resources.destroy is not None,
            )
        ):
            rows = (
                [["🟢", f"`{field}`"] for field in self.affected_resources.add]
                + [["🟡", f"`{field}`"] for field in self.affected_resources.change]
                + [["🔴", f"`{field}`"] for field in self.affected_resources.destroy]
            )
            return (
                tabulate(
                    rows,
                    headers=["Action", "Resources"],
                    colalign=("center",),
                    tablefmt="pipe",
                )
                if rows
                else ""
            )

        return "No affected resources"

    @property
    def _short_stdout(self):
        if self.stdout is None:
            return None
        output = decolor(self.stdout).splitlines()
        result_lines = []
        trigger_lines = [
            "Terraform has compared your real infrastructure against your configuration",
            "Terraform used the selected providers to generate the following execution",
        ]

        def _match(candidate):
            for trigger in trigger_lines:
                if candidate.startswith(trigger):
                    return True
            return False

        idx = 0
        # Find beginning of output we want to preserve
        while idx < len(output):
            if _match(output[idx]):
                break
            idx += 1

        # Save the rest of output
        while idx < len(output):
            result_lines.append(output[idx])
            if "~ user_data" in output[idx]:
                try:
                    parts = output[idx].split()
                    before = b64decode(parts[3].strip('"')).decode()
                    after = (
                        "(known after apply)"
                        if parts[5].strip('"') == "(known"
                        else b64decode(parts[5].strip('"')).decode()
                    )
                    result_lines.append("userdata changes:")
                    for diff_line in unified_diff(
                        before.splitlines(), after.splitlines(), fromfile="before", tofile="after", lineterm=""
                    ):
                        result_lines.append(diff_line)
                    result_lines.append("EOF userdata changes.")
                except UnicodeDecodeError as err:
                    LOG.warning("Failed to decode userdata: %s", err)
            idx += 1

        if result_lines:
            return "\n".join(result_lines)
        return "\n".join(output)

    def __eq__(self, other):
        return all(
            getattr(self, x) == getattr(other, x)
            for x in self.__dict__
            if x not in ["affected_resources", "stdout", "stderr"]
        )

    def __repr__(self):
        return json.dumps(
            {
                self.backend.id: {
                    "success": self.success,
                    "add": self.add,
                    "change": self.change,
                    "destroy": self.destroy,
                }
            }
        )



"""
.. topic:: ``ih-ec2 tags``

    A ``ih-ec2 tags`` subcommand.

    See ``ih-ec2 tags`` for more details.
"""

import json
from logging import getLogger

import click
from infrahouse_core.aws.ec2_instance import EC2Instance

from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.command(
    name="tags",
)
@click.argument("instance_id", required=False)
@click.pass_context
def cmd_tags(ctx, **kwargs):
    """
    List EC2 instance tags.

    By default, it will show tags of the local instance.

    """
    setup_logging(debug=ctx.obj["debug"])
    instance = EC2Instance(instance_id=kwargs["instance_id"], ec2_client=ctx.obj["ec2_client"])
    print(json.dumps(instance.tags, indent=4))



"""
.. topic:: ``ih-s3-reprepro``

    A ``ih-s3-reprepro`` command..

    See ``ih-s3-reprepro --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_check import cmd_check
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_checkpool import cmd_checkpool
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_deleteunreferenced import (
    cmd_deleteunreferenced,
)
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_dumpunreferenced import (
    cmd_dumpunreferenced,
)
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_get_secret_value import (
    cmd_get_secret_value,
)
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_includedeb import cmd_includedeb
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_list import cmd_list
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_migrate import cmd_migrate
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_remove import cmd_remove
from infrahouse_toolkit.cli.ih_s3_reprepro.cmd_set_secret_value import (
    cmd_set_secret_value,
)
from infrahouse_toolkit.cli.utils import DEPENDENCIES, check_dependencies
from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--bucket",
    help="AWS S3 bucket with a Debian repo.",
    required=True,
)
@click.option("--role-arn", help="Assume this role for all AWS operations.")
@click.option("--gpg-key-secret-id", help="AWS secrets manager secret name that stores a GPG private key.")
@click.option(
    "--gpg-passphrase-secret-id", help="AWS secrets manager secret name that stores a passphrase to the GPG key."
)
@click.option(
    "--aws-region",
    help="AWS region name. By default, let boto3 detect it.",
    default=None,
    type=click.Choice(
        [
            "us-east-2",
            "us-east-1",
            "us-west-1",
            "us-west-2",
            "af-south-1",
            "ap-east-1",
            "ap-south-2",
            "ap-southeast-3",
            "ap-southeast-4",
            "ap-south-1",
            "ap-northeast-3",
            "ap-northeast-2",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-northeast-1",
            "ca-central-1",
            "eu-central-1",
            "eu-west-1",
            "eu-west-2",
            "eu-south-1",
            "eu-west-3",
            "eu-south-2",
            "eu-north-1",
            "eu-central-2",
            "me-south-1",
            "me-central-1",
            "sa-east-1",
        ]
    ),
)
@click.pass_context
def ih_s3_reprepro(*args, **kwargs):  # pylint: disable=unused-argument
    """
    Tool to manage deb packages to a Debian repository hosted in an S3 bucket.
    """
    setup_logging(debug=kwargs["debug"])
    check_dependencies(DEPENDENCIES)


for cmd in [
    cmd_list,
    cmd_checkpool,
    cmd_check,
    cmd_remove,
    cmd_includedeb,
    cmd_dumpunreferenced,
    cmd_deleteunreferenced,
    cmd_set_secret_value,
    cmd_get_secret_value,
    cmd_migrate,
]:
    # noinspection PyTypeChecker
    ih_s3_reprepro.add_command(cmd)



"""
.. topic:: ``ih-ec2 list``

    A ``ih-ec2 list`` subcommand.

    See ``ih-ec2 list`` for more details.
"""

import json
import sys
from logging import getLogger
from pprint import pformat

import click
from botocore.exceptions import ClientError
from tabulate import tabulate

from infrahouse_toolkit.logging import setup_logging

LOG = getLogger(__name__)


def list_ec2_instances(ec2_client, fields=None, tag_filter=None, comma_separated=False):
    """
    Print a summary about EC2 instances in a region.
    """
    kwargs = {}
    if tag_filter:
        kwargs["Filters"] = tag_filter

    response = ec2_client.describe_instances(**kwargs)
    LOG.debug("describe_instances() = %s", pformat(response, indent=4))
    instances = []
    header = ["PrivateIpAddress", "InstanceId", "InstanceType"]
    fields = {} if fields is None else fields
    header.extend([k for k, v in fields.items() if v])
    for reservation in response["Reservations"]:
        for instance in reservation["Instances"]:
            name = None
            for tag in instance.get("Tags") or []:
                if tag["Key"] == "Name":
                    name = tag["Value"]
            row = [name]
            for field in header:
                value = instance[field] if field in instance else ""
                if field == "Tags":
                    value = json.dumps(
                        dict(
                            sorted(
                                {tag.get("Key"): tag.get("Value") for tag in value if tag.get("Key") != "Name"}.items()
                            )
                        ),
                        indent=4,
                    )
                row.append(value)
            row.append(instance["State"]["Name"])
            instances.append(row)

    if comma_separated:
        print(",".join([i[1] for i in instances]))
    else:
        print(
            tabulate(
                sorted(instances),
                headers=["Name"] + header + ["State"],
                tablefmt="grid" if fields["Tags"] else "outline",
            )
        )


@click.command(
    name="list",
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    },
)
@click.option(
    "--public-dns-name",
    "PublicDnsName",
    help="Show public DNS name.",
    is_flag=True,
    default=False,
)
@click.option(
    "--public-ip-address",
    "PublicIpAddress",
    help="Show public IP address.",
    is_flag=True,
    default=False,
)
@click.option(
    "-c",
    help="Output a comma-separated list instead of table.",
    is_flag=True,
    default=False,
)
@click.option(
    "--tags",
    "Tags",
    help="Show tags.",
    is_flag=True,
    default=False,
)
@click.pass_context
def cmd_list(ctx, **kwargs):
    """
    List created EC2 instances.

    By default, it will show instances' Name, PrivateIpAddress, InstanceId, InstanceType, and State.

    To display the instance's public DNS name or IP address, use options --public-dns-name,
    --public-ip-address respectively.

    Option --tags will show instance's tags.

    You can use tag names to filter output. For instance, option --service will show instances
    that have a tag 'service'. The same option with a value e.g. --service=vpn-portal will show instances
    that have a tag 'service' and its value 'vpn-portal'. Use comma separated service name to display more
    than one service. e.g. --service=elastic,elastic-kibana.
    """
    setup_logging(debug=ctx.obj["debug"], quiet=not kwargs["c"])

    ec2_client = ctx.obj["ec2_client"]
    aws_config = ctx.obj["aws_config"]
    tag_filter = []
    for arg in ctx.args:
        arg = arg.lstrip("--")
        split = arg.split("=")
        if len(split) < 2:
            name = "tag-key"
            values = [split[0]]
        else:
            name = f"tag:{split[0]}"
            values = list(split[1].split(","))

        tag_filter.append({"Name": name, "Values": values})

    try:
        comma_separated = kwargs.pop("c")
        list_ec2_instances(ec2_client, kwargs, tag_filter, comma_separated=comma_separated)
    except ClientError as err:
        LOG.exception(err)
        LOG.info("Try to run ih-ec2 with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)



"""
.. topic:: ``ih-elastic api``

    A ``ih-elastic api`` subcommand.

    See ``ih-elastic cluster-health --help`` for more details.
"""

import json
import os
import sys
from json import JSONDecodeError
from logging import getLogger
from os import path as osp

import click
import requests

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING

LOG = getLogger()


def _get_input(method, uri):
    ih_top = osp.expanduser("~/.infrahouse-toolkit")
    if not osp.exists(ih_top):
        os.mkdir(ih_top)
    tmpfile_name = osp.join(ih_top, f"ih-elastic-api-{method.lower()}-{''.join(e for e in uri if e.isalnum())}")
    click.edit(filename=tmpfile_name)
    os.chmod(tmpfile_name, 0o600)
    return open(tmpfile_name, encoding=DEFAULT_OPEN_ENCODING).read()


@click.command(name="api")
@click.option("-d", "--data", help="API call data.", is_flag=False, flag_value="editor", default=None)
@click.argument("METHOD", type=click.Choice(["GET", "HEAD", "POST", "PUT", "DELETE"]))
@click.argument("URI")
@click.pass_context
def cmd_api(ctx, **kwargs):
    """
    Execute a custom API call against the local Elasticsearch node.

    API documentation https://t.ly/-82Fw

    """
    method = kwargs["method"]
    url = ctx.obj["url"]
    uri = kwargs["uri"]
    request_kwargs = {"auth": ctx.obj["auth"], "headers": {"Content-Type": "application/json"}}
    if kwargs["data"] is not None:
        request_kwargs["data"] = _get_input(method, uri) if kwargs["data"] == "editor" else kwargs["data"]
        try:
            json.loads(request_kwargs["data"])
        except JSONDecodeError as err:
            LOG.error("Not a valid JSON:\n%s", request_kwargs["data"])
            LOG.error(err)
            sys.exit(1)

    full_url = f"{url}/{uri.lstrip('/')}"
    LOG.debug("Sending request %s", full_url)
    response = getattr(requests, method.lower())(full_url, **request_kwargs)
    try:
        print(json.dumps(response.json(), indent=4))
    except JSONDecodeError:
        print(response.text)



"""
.. topic:: ``ih-skeema run``

    A ``ih-skeema run`` subcommand.

    See ``ih-skeema run --help`` for more details.
"""

import logging
import sys
from subprocess import Popen

import click

LOG = logging.getLogger()

SKEEMA_COMMANDS = [
    "add-environment",
    "diff",
    "format",
    "help",
    "init",
    "lint",
    "pull",
    "push",
    "version",
]


@click.command(
    name="run",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@click.argument("skeema_command", type=click.Choice(SKEEMA_COMMANDS))
@click.pass_context
def cmd_run(ctx, *args, **kwargs):
    """
    Run a skeema command.
    """
    LOG.debug("args = %s", args)
    LOG.debug("kwargs = %s", kwargs)
    LOG.debug(ctx.args)
    cmd = [ctx.obj["skeema_path"], "--user", ctx.obj["username"], kwargs["skeema_command"]]
    cmd.extend(ctx.args)
    try:
        with Popen(cmd, env={"MYSQL_PWD": ctx.obj["password"]}) as proc:
            LOG.info("Launched command: %s", " ".join(cmd))
            proc.communicate()
            sys.exit(proc.returncode)

    except FileNotFoundError as err:
        LOG.error("Command `%s` failed to start.", " ".join(cmd))
        LOG.error("Please install Skeema first or specify a skeema executable via --skeema-path.")
        LOG.exception(err)
        sys.exit(1)



"""
.. topic:: ``gpg.py``

    GPG helper functions.
"""

from contextlib import contextmanager
from logging import getLogger
from os import path as osp
from subprocess import Popen
from tempfile import NamedTemporaryFile, TemporaryDirectory

from botocore.exceptions import ClientError

from infrahouse_toolkit import DEFAULT_ENCODING, DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.aws import get_client

LOG = getLogger()


@contextmanager
def gpg_home() -> str:
    """
    :return: GPG home directory
    """
    with TemporaryDirectory() as path:
        LOG.debug("GPG home is %s", path)
        yield path


@contextmanager
def gpg(secret_key=None, role_arn=None, secret_passphrase=None, region: str = None) -> str:
    """
    Prepare GPG keyring and yield path to it.
    If no function arguments are specified, it will return the default path ~/.gnupg.

    If secret_key is specified, the function will pull the GPG key from this secret.
    Optionally, it will assume a role, if given.

    If secret_passphrase isn't specified, gpg and reprepro will ask a passphrase from a terminal.
    If specified, the function will pull the passprase from AWS secretsmanager secret secret_passphrase,
    save it in a temporary file, and create a GPG config so gpg can read the passphrase from
    the temporary file.

    The GPG private key will be imported. Again, if secret_key is specified.
    So the GPG home will include the private key.

    :param secret_key: AWS secret id (name or ARN) with a GPG private key.
    :type secret_key: str
    :param role_arn: If specified, assume this role in AWS client.
    :type role_arn: str
    :param secret_passphrase: AWS secret id (name or ARN) with a passphrase for the GPG private key.
    :type secret_passphrase: str
    :param region: AWS region name.
    :type region: str
    :return: Path to GPG homedir.
    :rtype: str
    """
    if secret_key:
        LOG.debug("Reading GPG key from %s", secret_key)
        try:
            secrets_manager = get_client("secretsmanager", role_arn=role_arn, region=region)
            key = secrets_manager.get_secret_value(SecretId=secret_key)["SecretString"]
        except ClientError as err:
            LOG.error(err)
            if err.response["Error"]["Code"] == "ResourceNotFoundException":
                LOG.error("Make sure that you set a secret value to the key %s:", secret_key)
                LOG.error("ih-s3-reprepro ... set-secret-value %s /path/to/file/with/%s-value", secret_key, secret_key)
            raise

        with gpg_home() as homedir, NamedTemporaryFile() as gpg_key_desc, NamedTemporaryFile() as gpg_passphrase_desc:
            gpg_key_desc.write(key.encode(DEFAULT_ENCODING))
            gpg_key_desc.flush()

            cmd = ["gpg", "--homedir", homedir]

            if secret_passphrase:
                LOG.debug("Reading GPG key passphrase from %s", secret_passphrase)
                passphrase = secrets_manager.get_secret_value(SecretId=secret_passphrase)["SecretString"]
                gpg_passphrase_desc.write(passphrase.encode(DEFAULT_ENCODING))
                gpg_passphrase_desc.flush()
                write_gpg_cong(
                    osp.join(homedir, "gpg.conf"),
                    {"batch": None, "passphrase-file": gpg_passphrase_desc.name, "pinentry-mode": "loopback"},
                )

            cmd.extend(["--import", gpg_key_desc.name])
            LOG.debug("Executing: %s", " ".join(cmd))
            proc = Popen(cmd)
            proc.communicate()
            yield homedir
    else:
        yield osp.expanduser("~/.gnupg")


def write_gpg_cong(path: str, options: dict):
    """
    Prepare GPG config file.

    :param path: Path name to the GPG config file.
    :type path: str
    :param options: A dictionary with options. For one word options the key value is None.
    :type options: dict
    """
    LOG.debug("Saving GPG config in %s", path)
    with open(path, "w", encoding=DEFAULT_OPEN_ENCODING) as gpg_conf_desc:
        for key, value in options.items():
            gpg_conf_desc.write(f"{key}")
            if value:
                gpg_conf_desc.write(f" {value}")
            gpg_conf_desc.write("\n")



"""
.. topic:: ``ih-secrets get``

    A ``ih-secrets get`` subcommand.

    See ``ih-secrets get`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import ClientError

from infrahouse_toolkit.aws import get_secret

LOG = getLogger(__name__)


@click.command(name="get")
@click.argument("secret")
@click.pass_context
def cmd_get(ctx, secret):
    """
    Get a secret value.
    """
    secretsmanager_client = ctx.obj["secretsmanager_client"]
    aws_config = ctx.obj["aws_config"]
    try:
        print(get_secret(secretsmanager_client, secret))
    except ClientError as err:
        LOG.exception(err)
        LOG.error("Try to run ih-secrets with --aws-profile option.")
        LOG.error("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)



"""
.. topic:: ``utils.py``

    Various helper functions.
"""

import json
import os
import pathlib
import sys
from contextlib import contextmanager
from logging import getLogger
from os import getgid, getuid
from os import path as osp
from subprocess import PIPE, CalledProcessError, Popen, check_call, run
from tempfile import TemporaryDirectory
from time import sleep, time

import boto3

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.aws import (
    assume_role,
    get_credentials_from_environ,
    get_credentials_from_profile,
)
from infrahouse_toolkit.cli.gpg import gpg
from infrahouse_toolkit.exceptions import IHRetriableError

DEPENDENCIES = ["reprepro", "gpg", "s3fs"]
LOG = getLogger()


def check_dependencies(binaries: list):
    """
    Ensure that dependencies are installed. The function calls each of the binary
    with a ``--help`` option.

    :param binaries: Dependency commands. List of strings.
    :type binaries: list
    """
    for dep in binaries:
        try:
            with open("/dev/null", "w", encoding=DEFAULT_OPEN_ENCODING) as devnull:
                LOG.debug("Checking if %s is installed.", dep)
                check_call([dep, "--help"], stdout=devnull, stderr=devnull)
        except FileNotFoundError:
            LOG.error("Looks like %s is not installed", dep)
            LOG.info("Try installing it by \n\n\tapt-get install %s\n", dep)
            sys.exit(1)


def mount_s3(bucket: str, path: str, role_arn: str = None, region: str = None):
    """
    Mount an S3 bucket at a path.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param path: Local filesystem path name.
    :type path: str
    :param role_arn: Assume role if specified.
    :param region: AWS region name.
    :type region: str
    """
    LOG.debug("Mounting bucket %s on %s", bucket, path)
    env = {}
    cmd = ["s3fs", bucket, path, "-o", f"uid={getuid()}", "-o", f"gid={getgid()}"]
    if role_arn:
        LOG.debug("Using AWS credentials from IAM role %s", role_arn)
        env = assume_role(role_arn)
        sts = boto3.client(
            "sts",
            aws_access_key_id=env["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=env["AWS_SECRET_ACCESS_KEY"],
            aws_session_token=env["AWS_SESSION_TOKEN"],
            region_name=region,
        )
        response = sts.get_caller_identity()
        LOG.debug("Assumed role: %s", response)
    elif "AWS_ACCESS_KEY_ID" in os.environ:
        env = get_credentials_from_environ()
    else:
        env = get_credentials_from_profile()

    LOG.debug(
        "To reproduce environment: \n%s",
        "\n".join([f'export {key}="{value}"' for key, value in sanitize_env(env).items()]),
    )
    LOG.debug("Command to debug: mkdir -p %s; %s -o dbglevel=info -f -o curldbg", path, " ".join(cmd))
    execute(cmd, env=env)


def umount_s3(path: str):
    """
    Unmount an S3 bucket at a path.

    :param path: Local filesystem path name where the S3 bucket is mounted at.
    :type path: str
    """
    try:
        check_call(["umount", path])
    except CalledProcessError as err:
        LOG.exception(err)
        sys.exit(1)


@contextmanager
def local_s3(bucket, role_arn=None, retry_timeout=60, region=None, check_file="conf/distributions") -> str:
    """
    Mount an S3 bucket locally and return a mount point.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param role_arn: Assume role if specified.
    :type role_arn: str
    :param retry_timeout: How many second to keep trying to mount the bucket.
    :type retry_timeout: int
    :param check_file: When this file exists, the bucket is considered mounted.
    :type check_file: str
    :param region: AWS region name.
    :type region: str
    :return: Local filesystem path where the S3 bucket is mounted at.
    """
    with TemporaryDirectory() as mnt_dir:
        try:
            now = time()
            timeout = now + retry_timeout
            mount_s3(bucket, mnt_dir, role_arn=role_arn, region=region)
            while True:
                if osp.exists(osp.join(mnt_dir, check_file)):
                    LOG.debug("Mounted successfully %s on %s", bucket, mnt_dir)
                    break
                LOG.warning("Waiting until s3://%s is mounted at %s", bucket, mnt_dir)
                sleep(1)
                if time() > timeout:
                    raise RuntimeError(f"s3://{bucket} is not mounted after {retry_timeout} seconds")
            yield mnt_dir
        # except Exception as err:
        #     LOG.critical(err)
        #     raise

        finally:
            umount_s3(mnt_dir)


@contextmanager
def tmpfs_s3(bucket, role_arn=None, volume_size="512M") -> str:
    """
    Mount a temporary file system, sync an S3 bucket onto it.
    Then sync the local volume back to S3 an umount it.

    :param bucket: AWS S3 bucket name.
    :type bucket: str
    :param role_arn: Assume role if specified.
    :type role_arn: str
    :param volume_size: Temporary memory partition size. By default, 512M.
    :type volume_size: str
    :return: Local filesystem path where the S3 bucket is mounted at.
    """
    with TemporaryDirectory() as mnt_dir:
        try:
            run(
                ["mount", "-t", "tmpfs", "-o", f"size={volume_size}", "tmpfs", mnt_dir],
                capture_output=True,
                check=True,
            )
            # check_call(["aws", "s3", "sync", "--quiet", f"s3://{bucket}", mnt_dir])
            yield mnt_dir
            env = assume_role(role_arn) if role_arn else {}
            run(
                ["aws", "s3", "sync", "--only-show-errors", mnt_dir, f"s3://{bucket}"],
                capture_output=True,
                check=True,
                env={**env, **{"PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"}},
            )
        except CalledProcessError as err:
            LOG.error(err)
            LOG.error("stdout: %s", err.output)
            LOG.error("stderr: %s", err.stderr)
            raise
        finally:
            check_call(["umount", mnt_dir])


def execute(cmd: list, cwd: str = None, env: dict = None, exit_on_error: bool = True):
    """
    Execute a command and exit with 1 if the command raises CalledProcessError.

    :param cmd: A command to execute. It's passed to check_call() and therefore must be a list.
    :type cmd: list
    :param cwd: Working directory for the command.
    :type cwd: str
    :param env: Pass a dictionary with environment
    :type env: dict
    :param exit_on_error: If False, let the caller decide what to do on CalledProcessError exception.
    :type exit_on_error: bool
    """
    try:
        LOG.debug("Executing: '%s'", " ".join(cmd))
        LOG.debug("Environment: %s", json.dumps(sanitize_env(env), indent=4) if env else "None")
        with Popen(cmd, env=env, cwd=cwd, stdout=PIPE, stderr=PIPE) as proc:
            try:
                cout, cerr = proc.communicate()
                ret_code = proc.returncode
                LOG.debug("Command: '%s': Return code: %d", " ".join(cmd), ret_code)
                if cout:
                    print(cout.decode())
                if ret_code != 0:
                    print(cerr.decode())
                    raise CalledProcessError(
                        returncode=ret_code, cmd=" ".join(cmd), output=cout.decode(), stderr=cerr.decode()
                    )
            except KeyboardInterrupt:
                LOG.info("Exiting on <ctrl>+c")
                proc.terminate()
                LOG.info("Process '%s' is terminated. Waiting for it to exit.", cmd[0])
                proc.wait(60)

    except CalledProcessError as err:
        LOG.exception(err)
        LOG.error("CWD: %s", cwd)
        if exit_on_error:
            sys.exit(1)
        raise IHRetriableError(returncode=ret_code, cmd=cmd, output=err.output, stderr=err.stderr) from err
        # raise CalledProcessError(returncode=ret_code, cmd=" ".join(cmd)) from err


@contextmanager
def repo_env(bucket, role_arn, gpg_key_secret_id, gpg_passphrase_secret_id, region=None):
    """
    Prepare locally a repo and GPG so "reprepro" can manage it.

    :param bucket: AWS S3 bucket with the repo. The repo must be in the root.
    :type bucket: str
    :param role_arn: Optional role ARN. If specified, AWS client will assume it.
    :type role_arn: str
    :param gpg_key_secret_id: AWS secretsmanager secret (name or ARN) that stores the private GPG key,
        needed by ``reprepro`` to sign the repo.
    :type gpg_key_secret_id: str
    :param gpg_passphrase_secret_id: AWS secretsmanager secret (name or ARN) that stores a passphrase for
        the private GPG key. Note, it's not the passphrase itself, it's a secret that stores it.
    :type gpg_passphrase_secret_id: str
    :param region: AWS region name.
    :type region: str
    :return: A tuple with two strings: Local filesystem directory with a mounted S3 bucket
        and GPG home directory.
    """
    with local_s3(bucket, role_arn, region=region) as path:
        # pylint: disable-next=contextmanager-generator-missing-cleanup
        with gpg(
            secret_key=gpg_key_secret_id, role_arn=role_arn, secret_passphrase=gpg_passphrase_secret_id, region=region
        ) as gpg_home:
            yield path, gpg_home


def sanitize_env(env: dict = None) -> dict:
    """
    Mask secrets in environment variables.

    :param env: original environment
    :return: Environment without secret values
    """
    secret_keys = [
        "AWSSECRETACCESSKEY",
        "AWS_SECRET_ACCESS_KEY",
        "AWS_SESSION_TOKEN",
        "GITHUB_TOKEN",
        "GH_TOKEN",
        "AWSSESSIONTOKEN",
    ]
    return {key: value if key not in secret_keys else "***" for key, value in env.items()}


def mkdir_p(new_directory):
    """
    Equivalent of a shell command mkdir -p

    :param new_directory: Create a directory and its parents if needed.
    :type new_directory: str
    """
    pathlib.Path(new_directory).mkdir(parents=True, exist_ok=True)


def retry(func, args, kwargs, attempts: int = 5, retriable_exit_codes: list = None):
    """
    Execute a function and retry up to attempts times if it raised IHRetriableError

    :param func: Callable instance.
    :type func: callable
    :param args: Positional arguments to the function.
    :type args: tuple
    :param kwargs: Keyword arguments to the function.
    :type kwargs: dict
    :param attempts: Call the function up to this many times.
    :param retriable_exit_codes: IHRetriableError has a returncode. Retry only it the return code is in this list.
    """
    retriable_exit_codes = retriable_exit_codes or [1]
    sleep_time_sec = 1
    for attempt in range(attempts):
        try:
            return func(*args, **kwargs)
        except IHRetriableError as err:
            LOG.warning(err)
            LOG.warning("stdout: %s", err.output)
            LOG.warning("stderr: %s", err.stderr)
            if err.returncode in retriable_exit_codes:
                LOG.warning(
                    "Attempt %d out of %d failed. Will retry in %d second(s).", attempt, attempts, sleep_time_sec
                )
                sleep(sleep_time_sec)
                attempt += 1
                continue
            raise
    raise RuntimeError(f"Function didn't succeed after {attempts} attempts")



"""
.. topic:: ``ih-elastic cluster decommission-node``

    A ``ih-elastic cluster decommission-node`` subcommand.

    See ``ih-elastic cluster decommission-node --help`` for more details.
"""

import json
import sys
from logging import getLogger
from time import sleep

import click
from botocore.exceptions import ClientError
from elasticsearch.client import ClusterClient, NodesClient, ShutdownClient
from infrahouse_core.aws.asg import ASG
from infrahouse_core.aws.asg_instance import ASGInstance

from infrahouse_toolkit.lock.exceptions import LockAcquireError
from infrahouse_toolkit.lock.system import SystemLock
from infrahouse_toolkit.timeout import timeout

LOG = getLogger(__name__)


def wait_until_complete(client: ShutdownClient, node_id: str, wait_time: int = 3600, hook_name: str = "terminating"):
    """
    Give up to ``wait_time`` seconds to Elasticsearch to move shards out of the node.
    The function periodically checks the node's shutdown info
    (https://www.elastic.co/guide/en/elasticsearch/reference/current/get-shutdown.html)
    until it's ``COMPLETE``.
    If the node doesn't reach the ``COMPLETE`` state after ``wait_time`` seconds,
    raise TimeoutError.

    :param client: a shutdown Elasticsearch client.
    :type client: ShutdownClient
    :param node_id: Elasticsearch node identifier.
    :type node_id: str
    :param wait_time: Time in second to wait for the ``COMPLETE`` state.
    :type wait_time: int
    :param hook_name: Lifecycle hook name to extend while waiting.
    :type hook_name: str
    :raise TimeoutError: if after ``wait_time``, Elasticsearch hasn't moved all shards from the node.
    """
    local_instance = ASGInstance()
    asg = ASG(asg_name=local_instance.asg_name)
    if wait_time:
        try:
            with timeout(wait_time):
                while client.get_node(node_id=node_id).raw["nodes"][0]["status"] != "COMPLETE":
                    LOG.info(
                        "Current shutdown state:\n %s",
                        json.dumps(client.get_node(node_id=node_id).raw, indent=4),
                    )
                    if local_instance.lifecycle_state == "Terminating:Wait":
                        LOG.debug("Extend lifecycle hook %s", hook_name)
                        asg.record_lifecycle_action_heartbeat(hook_name=hook_name)

                    sleep(3)
        except TimeoutError as err:
            LOG.error(err)
            asg.cancel_instance_refresh()
            raise

    else:
        LOG.warning("wait_time = %r, not waiting for Elasticsearch to move shards out of this node.", wait_time)


def complete_lifecycle_action():
    """
    Completes the ``terminating`` lifecycle hook.
    If it fails, cancel all instance refreshes in the autoscaling group.
    """
    asg = ASG(asg_name=ASGInstance().asg_name)
    try:
        asg.complete_lifecycle_action()
    except ClientError as err:
        LOG.error(err)
        asg.cancel_instance_refresh()
        sys.exit(1)


@click.command(name="decommission-node")
@click.option(
    "--only-if-terminating",
    help="Proceed only if the instance is in a 'Terminating:Wait' lifecycle state.",
    default=None,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--wait-until-complete",
    help="Wait this many seconds until Elasticsearch completes moving shards out of this node.",
    default=3600,
    type=click.INT,
    show_default=True,
)
@click.option(
    "--complete-lifecycle-action",
    help="When it's safe, complete the lifecycle action.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--reason",
    help="Why the node is being decommissioned.",
    default=None,
    required=True,
    show_default=True,
)
@click.argument("node", required=False)
@click.pass_context
def cmd_decommission_node(ctx, **kwargs):
    """
    Prepares a node for decommissioning. Removes shards on of the node
    and can complete a lifecycle hook.

    If the NODE isn't specified, decommission the local node.
    """
    health = ClusterClient(ctx.obj["es"]).health()

    LOG.info("Checking %s cluster health: %s", health.body["cluster_name"], health.body["status"])

    if health.body["status"] != "green":
        LOG.info("The cluster status is %s - not green and, therefore, aborting.", health.body["status"])
        if kwargs["complete_lifecycle_action"]:
            ASG(asg_name=ASGInstance().asg_name).cancel_instance_refresh()
        sys.exit(1)

    node_id = list(NodesClient(ctx.obj["es"]).info(node_id="_local")["nodes"].keys())[0]

    only_if_terminating = kwargs["only_if_terminating"]
    local_instance = ASGInstance()
    shutdown_client = ShutdownClient(ctx.obj["es"])

    LOG.info("Current shutdown state:\n %s", json.dumps(shutdown_client.get_node(node_id=node_id).raw, indent=4))

    if only_if_terminating is None or (only_if_terminating and local_instance.lifecycle_state == "Terminating:Wait"):
        try:
            with SystemLock("/var/tmp/cmd_decommission_node.lock", blocking=False):
                LOG.info("Decommissioning node, %s", node_id)
                shutdown_client.put_node(node_id=node_id, type="remove", reason=kwargs["reason"])
                wait_until_complete(shutdown_client, node_id, wait_time=kwargs["wait_until_complete"])

                if kwargs["complete_lifecycle_action"]:
                    complete_lifecycle_action()

        except LockAcquireError as err:
            LOG.warning(err)
    else:
        LOG.info("No action performed.")
        LOG.info("--only-if-terminating %s", only_if_terminating)
        LOG.info("Lifecycle state: %s", local_instance.lifecycle_state)



"""
.. topic:: ``ih-elastic security api-key delete``

    A ``ih-elastic security api-key delete`` command.

    See ``ih-elastic security api-key delete --help`` for more details.
"""

import sys
from logging import getLogger

import click
from elasticsearch import NotFoundError
from elasticsearch.client import SecurityClient

LOG = getLogger(__name__)


@click.command(name="delete")
@click.argument("id")
@click.pass_context
def cmd_delete(ctx, **kwargs):
    """
    Create an API key.
    """
    client = SecurityClient(ctx.obj["es"])
    key_id = kwargs["id"]
    try:
        api_key = client.get_api_key(id=key_id)["api_keys"][0]
        client.invalidate_api_key(ids=[key_id])

        LOG.info(
            "API key %s:%s is invalidated. It will be fully deleted after the configured retention period.",
            api_key["id"],
            api_key["name"],
        )
    except NotFoundError:
        LOG.error("API key %s not found.", key_id)
        sys.exit(1)



"""
.. topic:: ``ih-elastic cat``

    A ``ih-elastic cat`` subcommand.

    See ``ih-elastic cat --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_elastic.cmd_cat.cmd_nodes import cmd_nodes
from infrahouse_toolkit.cli.ih_elastic.cmd_cat.cmd_repositories import cmd_repositories
from infrahouse_toolkit.cli.ih_elastic.cmd_cat.cmd_shards import cmd_shards
from infrahouse_toolkit.cli.ih_elastic.cmd_cat.cmd_snapshots import cmd_snapshots

LOG = getLogger(__name__)


@click.group(name="cat")
def cmd_cat():
    """
    Compact and aligned text (CAT) APIs.
    """


for cmd in [cmd_repositories, cmd_snapshots, cmd_shards, cmd_nodes]:
    # noinspection PyTypeChecker
    cmd_cat.add_command(cmd)



"""
cli module exceptions.
"""

from infrahouse_toolkit.cli.exceptions import IHCLIError


class IHGithubError(IHCLIError):
    """ih-github errors."""


class IHVariableNotFound(IHGithubError):
    """GitHub variable doesn't exist."""



"""
Module for EC2Instance class - a class tha represents an EC2 instance.
"""

from logging import getLogger

from cached_property import cached_property_with_ttl
from ec2_metadata import ec2_metadata

from infrahouse_toolkit.aws import get_client

LOG = getLogger()


class EC2Instance:
    """
    EC2Instance represents an EC2 instance.

    :param instance_id: Instance id. If omitted, the local instance is read from metadata.
    :type instance_id: str
    """

    def __init__(self, instance_id: str = None):
        self._instance_id = instance_id

    @property
    def availability_zone(self) -> str:
        """
        :return: Availability zone where this instance is hosted.
        """
        return ec2_metadata.availability_zone

    @property
    def instance_id(self) -> str:
        """
        :return: The instance's instance_id. It's read from metadata
            if the class instance was created w/o specifying it.
        """
        if self._instance_id is None:
            self._instance_id = ec2_metadata.instance_id
        return self._instance_id

    @property
    def state(self) -> str:
        """
        :return: EC2 instance state e.g. ``Running``, ``Terminated``, etc.
        """
        return self._describe_instance["State"]["Name"]

    @property
    def tags(self) -> dict:
        """
        :return: A dictionary with the instance tags. Keys are tag names, and values - the tag values.
        """
        return {tag["Key"]: tag["Value"] for tag in self._describe_instance["Tags"]}

    @property
    def _ec2_client(self):
        return get_client("ec2")

    @cached_property_with_ttl(ttl=10)
    def _describe_instance(self):
        return self._ec2_client.describe_instances(
            InstanceIds=[
                self.instance_id,
            ],
        )[
            "Reservations"
        ][0][
            "Instances"
        ][0]



"""
.. topic:: ``ih-aws autoscaling mark-unhealthy``

    A ``ih-aws autoscaling mark-unhealthy`` subcommand.

    See ``ih-aws autoscaling mark-unhealthy --help`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import ClientError

from infrahouse_toolkit.aws.asg_instance import ASGInstance

LOG = getLogger()


@click.command(name="mark-unhealthy")
@click.argument("instance_id", required=False)
def cmd_mark_unhealthy(**kwargs):
    """
    Mark instance Unhealthy so Autoscaling group will replace it.
    """
    try:
        ASGInstance(kwargs["instance_id"]).mark_unhealthy()

    except ClientError as err:
        LOG.error(err)
        sys.exit(1)



"""
.. topic:: ``ih-elastic security api-key``

    A ``ih-elastic security api-key`` subcommand.

    See ``ih-elastic security api-key --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_elastic.cmd_security.cmd_api_key.cmd_create import (
    cmd_create,
)
from infrahouse_toolkit.cli.ih_elastic.cmd_security.cmd_api_key.cmd_delete import (
    cmd_delete,
)
from infrahouse_toolkit.cli.ih_elastic.cmd_security.cmd_api_key.cmd_list import cmd_list

LOG = getLogger(__name__)


@click.group(name="api-key")
def cmd_api_key():
    """
    Work with API keys.
    """


for cmd in [cmd_list, cmd_create, cmd_delete]:
    # noinspection PyTypeChecker
    cmd_api_key.add_command(cmd)



"""
.. topic:: ``ih-elastic security``

    A ``ih-elastic security`` subcommand.

    See ``ih-elastic security --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_elastic.cmd_security.cmd_api_key import cmd_api_key

LOG = getLogger(__name__)


@click.group(name="security")
def cmd_security():
    """
    Security APIs.
    """


for cmd in [cmd_api_key]:
    # noinspection PyTypeChecker
    cmd_security.add_command(cmd)



"""
Module for ASGInstance class - a class to describe and work with
an instance that is a part of an Autoscaling group.
"""

from cached_property import cached_property_with_ttl

from infrahouse_toolkit.aws import get_client
from infrahouse_toolkit.aws.ec2_instance import EC2Instance


class ASGInstance(EC2Instance):
    """
    ASGInstance is an EC2 instance that is a part of an autoscaling group.
    Because it's an EC2 instance, ASGInstance inherits EC2Instance.
    """

    @property
    def lifecycle_state(self) -> str:
        """
        :return: Lifecycle state of the instance.
            See https://docs.aws.amazon.com/autoscaling/ec2/userguide/ec2-auto-scaling-lifecycle.html
            for possible values.
        """
        return self._describe_auto_scaling_instance["LifecycleState"]

    @property
    def asg_name(self) -> str:
        """
        :return: Name of an autoscaling group that this instance is a part of.
        """
        return self.tags["aws:autoscaling:groupName"]

    def mark_unhealthy(self):
        """Tell the autoscaling group that this instance is not healthy
        and should be replaced."""
        self._autoscaling_client.set_instance_health(
            InstanceId=self.instance_id,
            HealthStatus="Unhealthy",
        )

    @property
    def _autoscaling_client(self):
        return get_client("autoscaling")

    @cached_property_with_ttl(ttl=10)
    def _describe_auto_scaling_instance(self):
        return self._autoscaling_client.describe_auto_scaling_instances(
            InstanceIds=[
                self.instance_id,
            ],
        )[
            "AutoScalingInstances"
        ][0]



"""
AWS classes.
"""

import sys
import time
import webbrowser
from logging import getLogger
from os import environ
from os import path as osp
from pprint import pformat
from time import sleep

import boto3
import requests
from boto3 import Session
from botocore.exceptions import (
    ClientError,
    NoCredentialsError,
    SSOTokenLoadError,
    TokenRetrievalError,
)
from diskcache import Cache

from infrahouse_toolkit.aws.config import AWSConfig
from infrahouse_toolkit.aws.exceptions import IHAWSException
from infrahouse_toolkit.fs import ensure_permissions

VALUE_MAP = {
    "AWS_ACCESS_KEY_ID": "AccessKeyId",
    "AWS_SECRET_ACCESS_KEY": "SecretAccessKey",
    "AWS_SESSION_TOKEN": "SessionToken",
    # These are old s3.fs options
    # Soon they will be deprecated
    # https://github.com/s3fs-fuse/s3fs-fuse/pull/1729
    "AWSACCESSKEYID": "AccessKeyId",
    "AWSSECRETACCESSKEY": "SecretAccessKey",
    "AWSSESSIONTOKEN": "SessionToken",
}

LOG = getLogger()


def assume_role(role_arn, region=None) -> dict:
    """
    Assume a given role and return a dictionary with credentials.

    :param role_arn: Role to be assumed.
    :type role_arn: str
    :param region: AWS region name.
    :type region: str
    :return: A dictionary with three keys: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, and
    """
    try:
        LOG.debug("Assuming role %s", role_arn)
        client = boto3.client("sts", region_name=region)
        response = client.assume_role(RoleArn=role_arn, RoleSessionName="ih-s3-reprepro-s3fs")
        LOG.debug("Got credentials %r", response["Credentials"])
        return {var: response["Credentials"].get(key) for var, key in VALUE_MAP.items()}
    except ClientError as err:
        LOG.error(err)
        LOG.debug("To revert environment:\n%s", "\n".join([f"unset {key}" for key in VALUE_MAP]))
        raise


def get_aws_client(service_name: str, profile: str, region: str, session=None):
    """
    Get a client instance for an AWS service.

    :param service_name: AWS service e.g. ``ec2``.
    :param profile: AWS profile for authentication.
    :param region: AWS region.
    :param session: if an AWS session is passed, use it to create a client.
    :type session: Session
    :return: A client instance.
    """
    session = session or Session(region_name=region, profile_name=profile)
    return session.client(service_name)


def get_aws_session(aws_config: AWSConfig, aws_profile: str, aws_region: str) -> Session:
    """

    :param aws_config:
    :param aws_profile:
    :param aws_region:
    :return: Authenticated AWS session, or None if boto3 can connect to AWS without extra steps.
    """
    if aws_profile is None and "default" in aws_config.profiles:
        aws_profile = "default"

    try:
        response = get_aws_client("sts", aws_profile, aws_region).get_caller_identity()
        LOG.info("Connected to AWS as %s", response["Arn"])

    except (SSOTokenLoadError, TokenRetrievalError) as err:
        if not aws_profile:
            LOG.error("Try to run ih-aws with --aws-profile option.")
            LOG.error("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
            sys.exit(1)
        LOG.debug(err)
        aws_session = aws_sso_login(aws_config, aws_profile, region=aws_region)
        response = get_aws_client("sts", aws_profile, aws_region, session=aws_session).get_caller_identity()
        LOG.info("Connected to AWS as %s", response["Arn"])
        return aws_session

    except NoCredentialsError as err:
        LOG.error(err)
        LOG.info("Try to run ih-aws with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)

    return boto3.Session(region_name=aws_region)


def get_client(service_name, role_arn=None, region=None, session_name=__name__):
    """
    Get an AWS service client assuming a role if specified.

    :param service_name: AWS service. ec2, sts, etc.
    :type service_name: str
    :param role_arn: Role ARN if it needs to be assumed.
    :type role_arn: str
    :param session_name: A human-readable string that tells something about this session.
        Exact value isn't important.
    :type session_name: str
    :param region: AWS region name.
    :type region: str
    :return: AWS boto3 client.
    """
    if role_arn:
        sts = boto3.client("sts", region_name=region)
        aws_iam_role = sts.assume_role(RoleArn=role_arn, RoleSessionName=session_name)
        session = boto3.Session(
            aws_access_key_id=aws_iam_role["Credentials"]["AccessKeyId"],
            aws_secret_access_key=aws_iam_role["Credentials"]["SecretAccessKey"],
            aws_session_token=aws_iam_role["Credentials"]["SessionToken"],
            region_name=region,
        )
        sts = session.client("sts", region_name=region)
        response = sts.get_caller_identity()
        LOG.debug("Assumed role: %s", response)
        return session.client(service_name, region_name=region)

    return boto3.client(service_name, region_name=region)


def get_credentials_from_profile() -> dict:
    """
    Another way to get AWS credentials is from EC2 instance metadata.

    :return: A dictionary with AWS_* variables.
    """
    LOG.debug("Using AWS credentials from instance metadata")
    url = "http://169.254.169.254/latest/meta-data/iam/security-credentials/"
    profile_name = requests.get(url, timeout=10).text
    LOG.debug("Profile name %s", profile_name)
    profile_data = requests.get(f"{url}/{profile_name}", timeout=10).json()
    profile_data["SessionToken"] = profile_data["Token"]

    return {var: profile_data.get(key) for var, key in VALUE_MAP.items()}


def get_credentials_from_environ():
    """Yet another way to get credentials.

    If environment is already configured for AWS access, simply get the credential from the environment.
    This is a situation when a user configures AWS_* in their env.
    Or when a role has been assumed and AWS_* are configured.

    :return: A dictionary with AWS_* variables.
    """
    LOG.debug("Using AWS credentials from environment")
    return {
        "AWS_ACCESS_KEY_ID": environ.get("AWS_ACCESS_KEY_ID"),
        "AWS_SECRET_ACCESS_KEY": environ.get("AWS_SECRET_ACCESS_KEY"),
        "AWS_SESSION_TOKEN": environ.get("AWS_SESSION_TOKEN"),
        "AWSACCESSKEYID": environ.get("AWS_ACCESS_KEY_ID"),
        "AWSSECRETACCESSKEY": environ.get("AWS_SECRET_ACCESS_KEY"),
        "AWSSESSIONTOKEN": environ.get("AWS_SESSION_TOKEN"),
    }


def get_secret(secretsmanager_client, secret_name):
    """
    Retrieve a value of a secret by its name.
    """
    response = secretsmanager_client.get_secret_value(
        SecretId=secret_name,
    )
    LOG.debug("get_secrets() = %s", pformat(response, indent=4))
    return response["SecretString"]


def aws_sso_login(aws_config: AWSConfig, profile_name: str, region: str = None):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials
    """
    cache_directory = osp.expanduser("~/.infrahouse-toolkit")
    with Cache(directory=cache_directory) as cache_reference:
        cache_key = f"ih-ec2-credentials-{profile_name}"
        credentials = cache_reference.get(cache_key)
        if not credentials:
            credentials = _get_credentials(aws_config, profile_name)
            cache_reference.set(
                cache_key, credentials, expire=int(int(credentials["expiration"]) / 1000 - int(time.time()))
            )

    ensure_permissions(cache_directory, 0o700)
    return Session(
        region_name=region or aws_config.get_region(profile_name),
        aws_access_key_id=credentials["accessKeyId"],
        aws_secret_access_key=credentials["secretAccessKey"],
        aws_session_token=credentials["sessionToken"],
    )


def _get_credentials(aws_config: AWSConfig, profile_name: str):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials

    :raise IHAWSException: If user didn't confirm auth
    """

    session = Session()
    sso_oidc = session.client("sso-oidc")
    client_creds = sso_oidc.register_client(
        clientName="infrahouse-toolkit",
        clientType="public",
    )
    LOG.debug("client_creds = %s", pformat(client_creds, indent=4))
    device_authorization = sso_oidc.start_device_authorization(
        clientId=client_creds["clientId"],
        clientSecret=client_creds["clientSecret"],
        startUrl=aws_config.get_start_url(profile_name),
    )

    LOG.debug("device_authorization = %s", pformat(device_authorization, indent=4))
    url = device_authorization["verificationUriComplete"]
    device_code = device_authorization["deviceCode"]
    expires_in = device_authorization["expiresIn"]
    interval = device_authorization["interval"]
    LOG.info("Verify user code: %s", device_authorization["userCode"])
    webbrowser.open(url, autoraise=True)
    for _ in range(1, expires_in // interval + 1):
        sleep(interval)
        try:
            token = sso_oidc.create_token(
                grantType="urn:ietf:params:oauth:grant-type:device_code",
                deviceCode=device_code,
                clientId=client_creds["clientId"],
                clientSecret=client_creds["clientSecret"],
            )
            access_token = token["accessToken"]
            sso = session.client("sso")
            role_creds = sso.get_role_credentials(
                roleName=aws_config.get_role(profile_name),
                accountId=aws_config.get_account_id(profile_name),
                accessToken=access_token,
            )["roleCredentials"]
            return role_creds
        except sso_oidc.exceptions.AuthorizationPendingException:
            pass

    raise IHAWSException(f"The verification code isn't confirmed by user in {expires_in} seconds")



"""
.. topic:: ``ih-elastic cat nodes``

    A ``ih-elastic cat nodes`` subcommand.

    See ``ih-elastic cat nodes --help`` for more details.
"""

import json
from logging import getLogger

import click
from elasticsearch.client import CatClient, NodesClient

LOG = getLogger()


@click.command(name="nodes")
@click.argument("node_ip", required=False)
@click.pass_context
def cmd_nodes(ctx, **kwargs):
    """
    Lists nodes in a cluster.
    """
    if kwargs["node_ip"]:
        node_client = NodesClient(ctx.obj["es"])
        for node_id, node in node_client.info().raw["nodes"].items():
            if node["ip"] == kwargs["node_ip"]:
                LOG.info("Node %s", node_id)
                print(json.dumps(node, indent=4))

    else:
        client = CatClient(ctx.obj["es"])
        if ctx.obj["format"] == "json":
            print(json.dumps(client.nodes(format=ctx.obj["format"], v=True).raw, indent=4))
        else:
            print(client.nodes(format=ctx.obj["format"], v=True))



"""
.. topic:: ``ih-elastic passwd``

    A ``ih-elastic passwd`` subcommand.

    See ``ih-elastic passwd --help`` for more details.
"""

from logging import getLogger

import boto3
import click
from elasticsearch.client import SecurityClient

LOG = getLogger(__name__)


@click.command(name="passwd")
@click.option("--username", help="Username of whose the password will be changed.", required=True)
@click.option("--new-password-secret", help="AWS secretsmanager secret id with a new password.", required=True)
@click.pass_context
def cmd_passwd(ctx, **kwargs):
    """
    Change password for Elasticsearch user.
    """
    client = boto3.client("secretsmanager")
    response = client.get_secret_value(SecretId=kwargs["new_password_secret"])
    new_password = response["SecretString"]

    client = SecurityClient(ctx.obj["es"])
    client.change_password(username=kwargs["username"], password=new_password)
    LOG.info("Successfully changed password for user %s.", kwargs["username"])



"""
.. topic:: ``ih-s3-reprepro set-secret-value``

    A ``ih-s3-reprepro set-secret-value`` subcommand.

    See ``ih-s3-reprepro set-secret-value --help`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit.aws import get_client
from infrahouse_toolkit.cli.lib import read_from_file_or_prompt

LOG = getLogger()


@click.command(name="set-secret-value")
@click.argument("secret_id")
@click.argument("path", nargs=-1)
@click.pass_context
def cmd_set_secret_value(ctx: Context, secret_id, path):
    """
    Set value to a secret.

    Optionally the value may be given via a local file specified by a path argument.

    \b
    ih-s3-reprepro ... set-secret-value mysecret /path/to/file_with_value

    if the path is omitted, a user will be prompt for the value.
    """
    role_arn = ctx.parent.params["role_arn"]
    try:
        client = get_client("secretsmanager", role_arn=role_arn, region=ctx.parent.params["aws_region"])
        value = read_from_file_or_prompt(path[0])
        LOG.debug("Secret value: %s", value)
        client.put_secret_value(SecretId=secret_id, SecretString=value)
        LOG.info("Value of %s is successfully set.", secret_id)

    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



"""
.. topic:: ``ih-registry upload``

    A ``ih-registry upload`` subcommand.

    See ``ih-registry upload`` for more details.
"""

from logging import getLogger
from os import path as osp
from shutil import make_archive
from tempfile import TemporaryDirectory

import boto3
import click
import git
import requests

LOG = getLogger()


@click.command(name="upload")
@click.option(
    "--registry-url",
    help="InfraHouse Terraform Registry address",
    default="https://registry.infrahouse.com",
    show_default=True,
)
@click.option(
    "--dynamodb-table",
    help="Name of a DynamoDB table with deployment API keys",
    default="DeployKeys",
    show_default=True,
)
@click.option(
    "--deploy-key",
    help="Deployment API keys. By default, read from the DynamoDB table, specified by --dynamodb-table",
    default=None,
)
@click.option(
    "--module-name",
    help="Terraform module name in format <namespace>/<name>/<provider>. For example, infrahouse/cloud-init/aws",
)
@click.option(
    "--namespace",
    help="Terraform registry namespace. "
    "If --module-name isn't specified, this namespace will be used to construct the module name.",
    default="infrahouse",
    show_default=True,
)
@click.option(
    "--provider",
    help="Terraform provider name. "
    "If --module-name isn't specified, this provider name will be used to construct the module name.",
    default="aws",
    show_default=True,
)
@click.option(
    "--module-version",
    help="Terraform module name version. Must follow semantic versioning convention",
)
@click.option(
    "--module-path",
    help="Path to a directory with the Terraform module. The directory includes main.tf and other files",
    default=".",
    show_default=True,
)
def cmd_upload(**kwargs):
    """
    Upload Terraform module to the InfraHouse Terraform Registry
    """
    module_path = kwargs["module_path"]
    module_name = kwargs["module_name"] or "/".join(
        [
            kwargs["namespace"],
            _detect_module_name(module_path),
            kwargs["provider"],
        ]
    )
    module_version = kwargs["module_version"] or _detect_latest_tag(module_path)
    url = f"{kwargs['registry_url']}/terraform/modules/v1/{module_name}/{module_version}"
    LOG.debug("Uploading to %s", url)
    # 1 / 0
    api_key = kwargs["deploy_key"] or _detect_api_key(kwargs["dynamodb_table"], module_name)
    with TemporaryDirectory() as tmp_dir:
        LOG.debug("Archiving directory %s into directory %s", module_path, tmp_dir)
        archive_name = make_archive(osp.join(tmp_dir, "module"), "zip", kwargs["module_path"])
        LOG.debug("Archive %s", archive_name)
        with open(archive_name, "rb") as archive_desc:
            response = requests.post(
                url=url, headers={"x-api-key": api_key}, files={"archive": archive_desc.read()}, timeout=300
            )
            response.raise_for_status()
            LOG.info("Server response: code %d, body: %s", response.status_code, response.text or "empty")


def _detect_api_key(dynamodb_table: str, module_name: str) -> str:
    client = boto3.client("dynamodb")
    response = client.get_item(
        TableName=dynamodb_table,
        Key={
            "id": {
                "S": "-".join(module_name.split("/")),
            }
        },
    )
    return response["Item"]["key"]["S"]


def _detect_latest_tag(repo_path: str) -> str:
    repo = git.Repo(repo_path)
    tags = sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
    LOG.debug("tags in %s: %s", repo_path, tags)
    latest_tag = tags[-1]
    return latest_tag.name


def _detect_module_name(module_path: str) -> str:
    path = osp.basename(osp.abspath(module_path))
    path_parts = path.split("-")
    path_parts.pop(0)
    path_parts.pop(0)
    return "-".join(path_parts)



"""
.. topic:: ``ih-elastic snapshots create-repository``

    A ``ih-elastic snapshots create-repository`` subcommand.

    See ``ih-elastic snapshots create-repository --help`` for more details.
"""

from logging import getLogger

import click
from elasticsearch.client import SnapshotClient

LOG = getLogger(__name__)


@click.command(name="create-repository")
@click.argument("repository-name")
@click.argument("bucket-name")
@click.pass_context
def cmd_create_repository(ctx, **kwargs):
    """
    Creates a repository.
    """
    client = SnapshotClient(ctx.obj["es"])
    client.create_repository(
        name=kwargs["repository_name"],
        settings={
            "client": "default",
            "bucket": kwargs["bucket_name"],
        },
        type="s3",
    )



"""
.. topic:: ``ih-aws autoscaling``

    A ``ih-aws autoscaling`` subcommand.

    See ``ih-aws autoscaling --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_aws.cmd_autoscaling.cmd_mark_unhealthy import (
    cmd_mark_unhealthy,
)

LOG = getLogger()


@click.group(name="autoscaling")
def cmd_autoscaling():
    """
    AWS autoscaling Commands.
    """


for cmd in [
    cmd_mark_unhealthy,
]:
    # noinspection PyTypeChecker
    cmd_autoscaling.add_command(cmd)



"""
.. topic:: ``ih-elastic cat snapshots``

    A ``ih-elastic cat snapshots`` subcommand.

    See ``ih-elastic cat snapshots --help`` for more details.
"""

import sys
from logging import getLogger

import click
from elastic_transport import ConnectionTimeout
from elasticsearch.client import CatClient

LOG = getLogger(__name__)


@click.command(name="snapshots")
@click.pass_context
def cmd_snapshots(ctx):
    """
    Returns all snapshots.
    """
    try:
        client: CatClient = CatClient(ctx.obj["es"])
        kwargs = {"repository": "_all", "format": ctx.obj["format"], "v": True, "request_timeout": 3600}
        # Disabling unexpected-keyword-arg because in elasticsearch 8.12.0 package version pylint fails
        # However, actual test proves `request_timeout` to be a valid argument.
        # 8.15.0 fixes this problem however the package is backwards incompatible.
        LOG.info(
            "\n%s",
            client.snapshots(**kwargs),  # pylint: disable=unexpected-keyword-arg
        )
    except ConnectionTimeout as err:
        LOG.exception(err)
        LOG.error("try to pass --request-timeout")
        sys.exit(1)



"""
.. topic:: ``ih-s3-reprepro check``

    A ``ih-s3-reprepro check`` subcommand.

    See ``ih-s3-reprepro check --help`` for more details.
"""

import click
from click import Context

from infrahouse_toolkit.cli.utils import execute, local_s3


@click.command(name="check")
@click.argument("codenames", nargs=-1)
@click.pass_context
def cmd_check(ctx: Context, codenames):
    """Check for all needed files to be registered properly."""
    bucket = ctx.parent.params["bucket"]
    role_arn = ctx.parent.params["role_arn"]
    with local_s3(bucket, role_arn, region=ctx.parent.params["aws_region"]) as path:
        cmd = [
            "reprepro",
            "-V",
            "-b",
            path,
            "check",
        ]
        cmd.extend(codenames)
        execute(cmd)



"""
.. topic:: ``ih-certbot``

    A ``ih-certbot`` command, certbot wrapper

    See ``ih-certbot --help`` for more details.
"""

import sys
from logging import getLogger
from subprocess import CalledProcessError, check_call

import click

from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.command(
    "ih-certbot",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--quiet",
    help="Suppress informational messages and output only warnings and errors.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--certbot-command",
    help="Path to certbot command",
    default="/opt/infrahouse-toolkit/embedded/bin/certbot",
    show_default=True,
)
@click.pass_context
def ih_certbot(ctx, *args, **kwargs):  # pylint: disable=unused-argument
    """
    Tool to manage SSL certificates with the certbot client.
    It will run the certbot command packaged with the infrahouse-toolkit.

    To see the certbot's help run this:
    \b

    ih-certbot -- --help

    If infrahouse-toolkit is installed from a DEB package,
    the certbot command is also available in
    \b

    /opt/infrahouse-toolkit/embedded/bin/certbot
    """
    quiet = kwargs["quiet"]
    setup_logging(debug=kwargs["debug"], quiet=kwargs["quiet"])
    cmd = [kwargs["certbot_command"]]
    if quiet:
        cmd.append("--quiet")
    cmd.extend(ctx.args)
    LOG.debug("cmd = %s", cmd)
    try:
        with open("/dev/null", "wb") as devnull:
            check_call(cmd, stdout=devnull if quiet else None)
    except CalledProcessError as err:
        LOG.error(err)
        sys.exit(err.returncode)



"""
.. topic:: ``ih-s3 upload-logs``

    A ``ih-s3 upload-logs`` subcommand.

    See ``ih-s3 upload-logs`` for more details.
"""

import sys
from logging import getLogger
from subprocess import PIPE, Popen
from urllib.parse import urlparse

import click
from botocore.exceptions import ClientError
from infrahouse_core.aws.asg import ASG
from infrahouse_core.aws.asg_instance import ASGInstance

from infrahouse_toolkit.lock.exceptions import LockAcquireError
from infrahouse_toolkit.lock.system import SystemLock

LOG = getLogger()


def normalize_directory_name(source: str) -> str:
    """
    Remove slashes and dots from ends for the path and replace slashes with dashes.

    :param source: original path name.
    :return: modified path name.
    """
    return source.lstrip("/.").replace("/", "-")


@click.command(name="upload-logs")
@click.option(
    "--only-if-terminating",
    help="Proceed only if the instance is in a 'Terminating:Wait' lifecycle state.",
    default=None,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--wait-until-complete",
    help="Wait this many seconds until the logs are uploaded.",
    default=3600,
    type=click.INT,
    show_default=True,
)
@click.option(
    "--complete-lifecycle-action",
    help="Specify a lifecycle hook name to complete it.",
    default=None,
    show_default=True,
)
@click.argument("local_directory")
@click.argument("s3_path")
@click.pass_context
def cmd_upload_logs(ctx, **kwargs):
    """
    Archive the local directory and upload it to S3.
    For example, a command

    \b
    ih-s3 upload-logs /var/log s3://foo-bucket/path/to/

    will create an S3 object s3://foo-bucket/path/to/var-log.tar.gz
    """
    s3_client = ctx.obj["s3_client"]
    url_parts = urlparse(kwargs["s3_path"])
    local_instance = ASGInstance()
    only_if_terminating = kwargs["only_if_terminating"]

    if only_if_terminating is None or (only_if_terminating and local_instance.lifecycle_state == "Terminating:Wait"):
        try:
            with SystemLock("/var/tmp/cmd_upload_logs.lock", blocking=False):
                try:
                    cmd = ["tar", "zcf", "-", kwargs["local_directory"]]
                    with Popen(cmd, stdout=PIPE) as proc:
                        s3_client.upload_fileobj(
                            proc.stdout,
                            url_parts.netloc,
                            f"{url_parts.path.strip('/')}/{normalize_directory_name(kwargs['local_directory'])}.tar.gz",
                        )
                except ClientError as err:
                    LOG.exception(err)
                    sys.exit(1)
                if kwargs["complete_lifecycle_action"]:
                    ASG(asg_name=local_instance.asg_name).complete_lifecycle_action(kwargs["complete_lifecycle_action"])

        except LockAcquireError as err:
            LOG.warning(err)



"""
.. topic:: ``ih-aws``

    A group of commands to work with AWS.

    See ``ih-aws --help`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import NoRegionError

from infrahouse_toolkit.aws import get_aws_session
from infrahouse_toolkit.aws.config import AWSConfig
from infrahouse_toolkit.cli.ih_aws.cmd_autoscaling import cmd_autoscaling
from infrahouse_toolkit.cli.ih_aws.cmd_credentials import cmd_credentials
from infrahouse_toolkit.cli.ih_aws.cmd_ecs import cmd_ecs
from infrahouse_toolkit.logging import setup_logging

AWS_DEFAULT_REGION = "us-west-1"
LOG = getLogger(__name__)


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--verbose",
    help="Print INFO and WARNING level messages.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--aws-profile",
    help="AWS profile name for authentication.",
    type=click.Choice(AWSConfig().profiles),
    default=None,
    show_default=True,
)
@click.option(
    "--aws-region",
    help="AWS region to use.",
    type=click.Choice(AWSConfig().regions),
    show_default=True,
    default=None,
)
@click.version_option()
@click.pass_context
def ih_aws(ctx, **kwargs):
    """AWS helpers."""
    setup_logging(debug=kwargs["debug"], quiet=not kwargs["verbose"])
    aws_profile = kwargs["aws_profile"]
    aws_region = kwargs["aws_region"]
    aws_config = AWSConfig()

    try:
        ctx.obj = {
            "debug": kwargs["debug"],
            "aws_session": get_aws_session(aws_config, aws_profile, aws_region),
            "aws_config": aws_config,
            "aws_profile": aws_profile,
        }
    except NoRegionError as err:
        LOG.error(err)
        LOG.error("Use the --aws-region option to specify the AWS region.")
        sys.exit(1)


for cmd in [
    cmd_credentials,
    cmd_ecs,
    cmd_autoscaling,
]:
    # noinspection PyTypeChecker
    ih_aws.add_command(cmd)



"""
.. topic:: ``ih-s3``

    A group of commands to work with AWS S3 service.

    See ``ih-s3 --help`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import NoRegionError
from infrahouse_core.aws import AWSConfig, get_aws_client, get_aws_session

from infrahouse_toolkit.cli.ih_s3.cmd_upload_logs import cmd_upload_logs
from infrahouse_toolkit.logging import setup_logging

AWS_DEFAULT_REGION = "us-west-1"
LOG = getLogger(__name__)


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--aws-profile",
    help="AWS profile name for authentication.",
    type=click.Choice(AWSConfig().profiles),
    default=None,
    show_default=True,
)
@click.option(
    "--aws-region",
    help="AWS region to use.",
    type=click.Choice(AWSConfig().regions),
    show_default=True,
    default=None,
)
@click.version_option()
@click.pass_context
def ih_s3(ctx, **kwargs):
    """AWS S3 helpers."""
    setup_logging(debug=kwargs["debug"])

    aws_profile = kwargs["aws_profile"]
    aws_config = AWSConfig()
    aws_region = kwargs["aws_region"]
    aws_session = get_aws_session(aws_config, aws_profile, aws_region)

    try:
        ctx.obj = {
            "debug": kwargs["debug"],
            "s3_client": get_aws_client("s3", aws_profile, aws_region, session=aws_session),
            "aws_config": aws_config,
        }
    except NoRegionError as err:
        LOG.error(err)
        LOG.error("Use the --aws-region option to specify the AWS region.")
        sys.exit(1)


for cmd in [cmd_upload_logs]:
    # noinspection PyTypeChecker
    ih_s3.add_command(cmd)



"""
.. topic:: ``ih-github runner check-health``

    A ``ih-github runner check-health`` subcommand.

    See ``ih-github runner check-health --help`` for more details.
"""

import logging
import sys
from shutil import disk_usage
from subprocess import DEVNULL, CalledProcessError, check_call

import click

from infrahouse_toolkit.aws.asg_instance import ASGInstance

LOG = logging.getLogger()


@click.command(
    name="check-health",
)
@click.option(
    "--disk-usage-threshold",
    help="How much the disk can be used in %% before the instance is considered healthy.",
    default=99.0,
    show_default=True,
)
def cmd_check_health(**kwargs):
    """
    Check if runner is online and healthy. If not healthy, make the instance as Unhealthy
    in the autoscaling group.

    Exit code is zero if healthy. Otherwise, 1.
    """
    checks = {_disk_usage: {"threshold": kwargs["disk_usage_threshold"]}, _check_is_service_running: {}}
    for func, keywargs in checks.items():
        if not func(**keywargs):
            ASGInstance().mark_unhealthy()
            sys.exit(1)


def _disk_usage(path="/", threshold=99.0) -> bool:
    stat = disk_usage(path)
    du_pct = 100 * stat.used / stat.total
    if du_pct > threshold:
        LOG.error("Disk usage on partition %s is %f%%, more than threshold %f%%", path, du_pct, threshold)
        return False
    LOG.debug("Disk usage on partition %s is %f%%", path, du_pct)
    return True


def _check_is_service_running(service_name="actions-runner") -> bool:
    try:
        check_call(
            ["systemctl", "is-active", service_name],
            stdout=DEVNULL,
        )
        return True
    except CalledProcessError as err:
        LOG.error("Error checking service status: %s", err)
        return False



"""
.. topic:: ``ih-github runner deregister``

    A ``ih-github runner deregister`` subcommand.

    See ``ih-github runner deregister --help`` for more details.
"""

import logging

import click
from requests import delete

LOG = logging.getLogger()


@click.command(
    name="deregister",
)
@click.argument("runner_id", type=click.INT)
@click.pass_context
def cmd_deregister(ctx, **kwargs):
    """
    deregister a self-hosted runner.
    """
    github_token = ctx.obj["github_token"]
    org = ctx.obj["org"]

    response = delete(
        f"https://api.github.com/orgs/{org}/actions/runners/{kwargs['runner_id']}",
        headers={
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {github_token}",
            "X-GitHub-Api-Version": "2022-11-28",
        },
        timeout=30,
    )
    response.raise_for_status()
    LOG.info("Runner %d is successfully deregistered.", kwargs["runner_id"])



"""
.. topic:: ``ih-ec2 launch``

    A ``ih-ec2 launch`` subcommand.

    See ``ih-ec2 launch`` for more details.
"""

import sys
from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_ec2.cmd_launch_templates import list_launch_templates
from infrahouse_toolkit.cli.ih_ec2.cmd_subnets import list_subnets
from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()
SUPPORTED_UBUNTU_CODENAMES = ["jammy"]


def launch_ec2_instance(ec2_client, kwargs):
    """
    Launch EC2 instance.
    """
    response = ec2_client.run_instances(**kwargs)
    LOG.info("Successfully started instance %s.", response["Instances"][0]["InstanceId"])


@click.command(name="launch")
@click.option(
    "--subnet-id",
    help="Subnet ID where to launch the instance.",
    default=None,
)
@click.argument("launch_template", required=False)
@click.pass_context
def cmd_launch(ctx, subnet_id, launch_template):
    """
    Start an EC2 instance.
    """
    setup_logging(debug=ctx.obj["debug"])
    ec2_client = ctx.obj["ec2_client"]
    if not subnet_id:
        LOG.error("Please specify --subnet-id from following:")
        list_subnets(ec2_client)
        sys.exit(1)

    if launch_template:
        LOG.info("Using launch template %s", launch_template)
        kwargs = {"MinCount": 1, "MaxCount": 1, "LaunchTemplate": {"Version": "$Latest"}, "SubnetId": subnet_id}
        kwargs["LaunchTemplate"][
            "LaunchTemplateId" if launch_template.startswith("lt-") else "LaunchTemplateName"
        ] = launch_template
        launch_ec2_instance(ctx.obj["ec2_client"], kwargs)

    else:
        LOG.error(
            "A launch template isn't specified. "
            "Please pick one from the list and pass either the launch template name or id."
        )
        list_launch_templates(ec2_client)



"""
.. topic:: ``ih-elastic security api-key create``

    A ``ih-elastic security api-key create`` command.

    See ``ih-elastic security api-key create --help`` for more details.
"""

import json
from logging import getLogger

import click
from elasticsearch.client import SecurityClient

LOG = getLogger(__name__)


@click.command(name="create")
@click.argument("name")
@click.pass_context
def cmd_create(ctx, name):
    """
    Create an API key.
    """
    client = SecurityClient(ctx.obj["es"])
    api_key = client.create_api_key(name=name)
    LOG.info("API key %s is created.", name)
    print(json.dumps(dict(api_key.body), indent=4))



"""
.. topic:: ``ih-github backup``

    A ``ih-github backup`` subcommand.

    See ``ih-github backup --help`` for more details.
"""

import logging
from datetime import datetime
from multiprocessing import Pool, set_start_method
from os import path as osp
from shutil import rmtree
from urllib.parse import urlparse

import click
import requests
from github import GithubIntegration
from github.Consts import MAX_JWT_EXPIRY

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.aws import get_client, get_secret
from infrahouse_toolkit.cli.ih_github.exceptions import IHVariableNotFound
from infrahouse_toolkit.cli.utils import (
    check_dependencies,
    execute,
    mkdir_p,
    retry,
    tmpfs_s3,
)
from infrahouse_toolkit.logging import setup_logging

LOG = logging.getLogger()
GH_APP_ID = "1016509"


def _get_variable(name, gh_token, org_name):
    url = f"https://api.github.com/orgs/{org_name}/actions/variables/{name}"
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {gh_token}",
        "X-GitHub-Api-Version": "2022-11-28",
    }
    response = requests.get(url, headers=headers, timeout=600)
    if response.status_code == 404:
        raise IHVariableNotFound(f"Variable {name} doesn't exist.")

    return response.json()["value"]


def _get_backup_bucket(gh_token, org_name):
    return _get_variable("INFRAHOUSE_BACKUP_BUCKET", gh_token, org_name)


def _get_backup_role(gh_token, org_name):
    return _get_variable("INFRAHOUSE_BACKUP_ROLE", gh_token, org_name)


def _get_org_name(github_client, installation_id):
    url = f"https://api.github.com/app/installations/{installation_id}"
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {github_client.create_jwt()}",
        "X-GitHub-Api-Version": "2022-11-28",
    }
    response = requests.get(url, headers=headers, timeout=600)
    return response.json()["account"]["login"]


@click.command(
    name="backup",
)
@click.option(
    "--app-key-url",
    help="URL to fetch the GitHub App key. Supports schemes file://, secretsmanager://",
    default=None,
    required=True,
)
@click.option(
    "--only-installation-id",
    help="Backup only a specified installation id.",
    type=click.INT,
)
@click.option(
    "--tmp-volume-size",
    help="Size of in-memory temporary file system.",
    default="512M",
    show_default=True,
)
@click.pass_context
def cmd_backup(ctx, **kwargs):
    """
    infrahouse-github-backup implementation.

    For each GitHub App installation,

    * Reads configuration from the Org: IAM role, S3 bucket, list of repos
    * Assumes, the IAM role
    * Mounts the S3 Bucket
    * Syncs each repo to the S3 bucket.
    """
    check_dependencies(["s3fs"])
    app_key_url = urlparse(kwargs["app_key_url"])
    LOG.debug("GitHub App key URL: %r", app_key_url)
    if app_key_url.scheme == "file":
        with open(f"/{app_key_url.netloc}{app_key_url.path}", "r", encoding=DEFAULT_OPEN_ENCODING) as key_file:
            app_key = key_file.read()
    elif app_key_url.scheme == "secretsmanager":
        app_key = get_secret(get_client("secretsmanager"), app_key_url.netloc)
    else:
        raise RuntimeError(f"Unsupported App key URL scheme {app_key_url.scheme}")

    github_client = GithubIntegration(GH_APP_ID, app_key, jwt_expiry=MAX_JWT_EXPIRY)
    set_start_method("spawn")
    for installation in github_client.get_installations():
        LOG.info("Processing installation %r", installation)
        if installation.target_type != "Organization":
            LOG.warning(
                "Installation %d is type %s, which is not Organization. It's not supported.",
                installation.id,
                installation.target_type,
            )
            continue
        org_name = _get_org_name(github_client, installation.id)
        if kwargs["only_installation_id"] and kwargs["only_installation_id"] != installation.id:
            LOG.warning(
                "Skipping installation id %d (%s) because --only-installation-id %d was requested.",
                installation.id,
                org_name,
                kwargs["only_installation_id"],
            )
            continue
        token = github_client.get_access_token(installation_id=installation.id).token
        try:
            bucket_name = _get_backup_bucket(token, org_name)
            with tmpfs_s3(
                bucket_name, role_arn=_get_backup_role(token, org_name), volume_size=kwargs["tmp_volume_size"]
            ) as path:
                with Pool() as pool:
                    pool.starmap(
                        _backup_repo, [(repo, path, token, ctx.obj["debug"]) for repo in installation.get_repos()]
                    )
            LOG.info("Backing up installation %r is done.", installation)

        except IHVariableNotFound as err:
            LOG.error(err)
            LOG.error("Organization %s is not fully configured.", org_name)


def _backup_repo(repository, dst_path, token, debug=False):
    setup_logging(debug=debug)
    LOG.debug("Backing up repository %s", repository)
    mkdir_p(osp.join(dst_path, str(repository.owner.login)))
    suffix = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    retry(
        execute,
        (
            [
                "gh",
                "repo",
                "clone",
                repository.full_name,
                osp.join(repository.full_name, suffix),
                "--",
                "--mirror",
            ],
        ),
        {
            "cwd": dst_path,
            "env": {
                "PATH": "/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:",
                "GH_TOKEN": token,
            },
            "exit_on_error": False,
        },
    )
    execute(
        ["tar", "zcf", f"{suffix}.tar.gz", suffix],
        cwd=osp.join(dst_path, str(repository.full_name)),
        env={
            "PATH": "/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:",
        },
    )
    rmtree(osp.join(dst_path, str(repository.full_name), suffix))



from os import environ

import pytest
from github import Github, InputFileContent


@pytest.mark.skipif("GITHUB_TOKEN" not in environ, reason="This is a development test, needs GITHUB_TOKEN")
def test_publish_gist():
    gh = Github(login_or_token=environ["GITHUB_TOKEN"])
    user = gh.get_user()
    gist = user.create_gist(public=False, files={"test": InputFileContent("foo", "new_name")})
    print(gist.html_url)



"""
.. topic:: ``ih-registry``

    A group of commands to work with InfraHouse Terraform registry.

    See ``ih-registry --help`` for more details.
"""

from logging import getLogger

import click

from infrahouse_toolkit.cli.ih_registry.cmd_upload import cmd_upload
from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.version_option()
@click.pass_context
def ih_registry(ctx, **kwargs):
    """InfraHouse Terraform Registry helpers."""
    setup_logging(debug=kwargs["debug"])
    ctx.obj = {"debug": kwargs["debug"]}


for cmd in [cmd_upload]:
    # noinspection PyTypeChecker
    ih_registry.add_command(cmd)



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

CODEC = "utf-8"

with open("README.rst", encoding=CODEC) as readme_file:
    readme = readme_file.read()

HISTORY = "See git log"


def parse_requirements(req_file):
    """
    Parse file with requirements and return a dictionary
    consumable by setuptools.

    :param req_file: path to requirements file.
    :type req_file: str
    :return: Dictionary with requirements.
    :rtype: dict
    """
    with open(req_file, encoding=CODEC) as f_descr:
        reqs = f_descr.read().strip().split("\n")
    return [x for x in reqs if x and not x.strip().startswith("#")]


requirements = parse_requirements("requirements.txt")
test_requirements = parse_requirements("requirements_dev.txt")
console_scripts = [
    "ih-aws=infrahouse_toolkit.cli.ih_aws:ih_aws",
    "ih-certbot=infrahouse_toolkit.cli.ih_certbot:ih_certbot",
    "ih-ec2=infrahouse_toolkit.cli.ih_ec2:ih_ec2",
    "ih-elastic=infrahouse_toolkit.cli.ih_elastic:ih_elastic",
    "ih-github=infrahouse_toolkit.cli.ih_github:ih_github",
    "ih-plan=infrahouse_toolkit.cli.ih_plan:ih_plan",
    "ih-puppet=infrahouse_toolkit.cli.ih_puppet:ih_puppet",
    "ih-registry=infrahouse_toolkit.cli.ih_registry:ih_registry",
    "ih-s3=infrahouse_toolkit.cli.ih_s3:ih_s3",
    "ih-s3-reprepro=infrahouse_toolkit.cli.ih_s3_reprepro:ih_s3_reprepro",
    "ih-secrets=infrahouse_toolkit.cli.ih_secrets:ih_secrets",
    "ih-skeema=infrahouse_toolkit.cli.ih_skeema:ih_skeema",
]

setup(
    author="Oleksandr Kuzminskyi",
    author_email="aleks@infrahouse.com",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="A collection of tools for building infrastructure.",
    entry_points={
        "console_scripts": console_scripts,
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + HISTORY,
    include_package_data=True,
    keywords="infrahouse-toolkit",
    name="infrahouse-toolkit",
    packages=find_packages(include=["infrahouse_toolkit", "infrahouse_toolkit.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/infrahouse/infrahouse-toolkit",
    version="2.51.0",
    zip_safe=False,
)



"""
.. topic:: ``ih-elastic security api-key list``

    A ``ih-elastic security api-key list`` command.

    See ``ih-elastic security api-key list --help`` for more details.
"""

from logging import getLogger

import click
from elasticsearch.client import SecurityClient
from tabulate import tabulate

LOG = getLogger(__name__)


@click.command(name="list")
@click.pass_context
def cmd_list(ctx):
    """
    List API keys.
    """
    client = SecurityClient(ctx.obj["es"])
    header = [
        "id",
        "name",
        "type",
        "creation",
        "invalidated",
        "username",
        "realm",
        "realm_type",
        "metadata",
        "role_descriptors",
    ]
    api_keys = []
    for api_key in client.get_api_key()["api_keys"]:
        row = [api_key[h] for h in header]
        api_keys.append(row)

    print(tabulate(api_keys, headers=header, tablefmt="outline"))



"""
.. topic:: ``ih-ec2 subnets``

    A ``ih-ec2 subnets`` subcommand.

    See ``ih-ec2 subnets`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import ClientError
from tabulate import tabulate

from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


def get_vpc_name(ec2_client, vpc_id):
    """
    Given a vpc_id find out its name.
    """
    response = ec2_client.describe_vpcs(
        VpcIds=[
            vpc_id,
        ]
    )
    for tag in response["Vpcs"][0]["Tags"]:
        if tag["Key"] == "Name":
            return tag["Value"]

    return vpc_id


def list_subnets(ec2_client):
    """
    Find and print information about available subnets in a region.
    """
    response = ec2_client.describe_subnets()
    header = ["SubnetId", "Name", "Public", "CidrBlock", "VpcId", "VpcName"]
    rows = []
    vpc_names = {}
    for subnet in response["Subnets"]:
        row = []
        vpc_id = subnet["VpcId"]
        for field in header:
            if field == "Name":
                value = ""
                for tag in subnet["Tags"]:
                    if tag["Key"] == "Name":
                        value = tag["Value"]
            elif field == "VpcName":
                if vpc_id not in vpc_names:
                    vpc_names[vpc_id] = get_vpc_name(ec2_client, vpc_id)
                value = vpc_names[vpc_id]
            elif field == "Public":
                value = subnet["MapPublicIpOnLaunch"]
            else:
                value = subnet[field]
            row.append(value)
        rows.append(row)

    print(tabulate(sorted(rows, key=lambda x: x[1]), headers=header, tablefmt="outline"))


@click.command(name="subnets")
@click.pass_context
def cmd_subnets(ctx):
    """
    Describe AWS subnets.
    """
    setup_logging(debug=ctx.obj["debug"])
    ec2_client = ctx.obj["ec2_client"]
    aws_config = ctx.obj["aws_config"]
    try:
        list_subnets(ec2_client)
    except ClientError as err:
        LOG.exception(err)
        LOG.info("Try to run ih-ec2 with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)



"""
.. topic:: ``ih-github runner register``

    A ``ih-github runner register`` subcommand.

    See ``ih-github runner register --help`` for more details.
"""

import logging
from os import path as osp
from subprocess import run

import boto3
import click
from requests import post

from infrahouse_toolkit.cli.ih_secrets.cmd_get import get_secret

LOG = logging.getLogger()
AR_URL = "https://github.com/actions/runner/releases/download/v2.314.1/actions-runner-linux-x64-2.314.1.tar.gz"


@click.command(
    name="register",
)
@click.option(
    "--actions-runner-code-path",
    help=f"Path to a directory with the actions-runner code. You can download it from {AR_URL}",
    show_default=True,
    default="/tmp/actions-runner-linux",
)
@click.option(
    "--label",
    help="Add a label to the runner.",
    multiple=True,
)
@click.argument("URL")
@click.pass_context
def cmd_register(ctx, **kwargs):
    """
    register a self-hosted runner. A given URL can be either org or a repo address.

    """
    github_token = ctx.obj["github_token"]
    org = ctx.obj["org"]
    registration_token_secret = ctx.obj["registration_token_secret"]

    if registration_token_secret:
        token = get_secret(boto3.client("secretsmanager"), registration_token_secret)

    else:
        response = post(
            f"https://api.github.com/orgs/{org}/actions/runners/registration-token",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {github_token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
            timeout=30,
        )
        response.raise_for_status()
        token = response.json()["token"]

    cmd = [
        osp.join(kwargs["actions_runner_code_path"], "config.sh"),
        "--url",
        kwargs["url"],
        "--token",
        token,
        "--unattended",
        "--disableupdate",
    ]
    if kwargs["label"]:
        cmd.extend(["--labels", ",".join(kwargs["label"])])
    LOG.debug("Executing %s", " ".join(cmd))
    run(
        cmd,
        check=True,
        env={
            "RUNNER_ALLOW_RUNASROOT": "microsoft_sux",
            "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
        },
    )



"""
.. topic:: ``ih-elastic``

    A ``ih-elastic`` command, Elasticsearch helper.

    See ``ih-elastic --help`` for more details.
"""

import socket
import sys
from logging import getLogger

import boto3
import click
from elasticsearch import Elasticsearch
from requests.auth import HTTPBasicAuth

from infrahouse_toolkit.cli.ih_elastic.cmd_api import cmd_api
from infrahouse_toolkit.cli.ih_elastic.cmd_cat import cmd_cat
from infrahouse_toolkit.cli.ih_elastic.cmd_cluster import cmd_cluster
from infrahouse_toolkit.cli.ih_elastic.cmd_cluster_health import cmd_cluster_health
from infrahouse_toolkit.cli.ih_elastic.cmd_passwd import cmd_passwd
from infrahouse_toolkit.cli.ih_elastic.cmd_security import cmd_security
from infrahouse_toolkit.cli.ih_elastic.cmd_snapshots import cmd_snapshots
from infrahouse_toolkit.cli.lib import get_elastic_password
from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


@click.group(
    "ih-elastic",
)
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--quiet",
    help="Suppress informational messages and output only warnings and errors.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--username",
    help="Username in Elasticsearch cluster.",
    default="elastic",
    show_default=True,
)
@click.option(
    "--password",
    help="Password for the Elasticsearch user. By default try to read it from puppet facts/AWS secretsmanager.",
    default=None,
    show_default=False,
)
@click.option(
    "--password-secret", help="AWS secretsmanager secret id with the password.", default=None, show_default=True
)
@click.option("--es-protocol", help="Elasticsearch protocol", default="http", show_default=True)
@click.option(
    "--es-host", help="Elasticsearch host", default=socket.gethostbyname(socket.gethostname()), show_default=True
)
@click.option("--es-port", help="Elasticsearch port", default=9200, show_default=True)
@click.option("--format", help="Output format", type=click.Choice(["text", "json", "cbor", "yaml", "smile"]))
@click.option(
    "--request-timeout", help="Timeout for HTTP requests in seconds.", default=10, show_default=True, type=click.INT
)
@click.version_option()
@click.pass_context
def ih_elastic(ctx, **kwargs):  # pylint: disable=unused-argument
    """
    Elasticsearch helper.
    """
    setup_logging(debug=kwargs["debug"], quiet=kwargs["quiet"])
    client = boto3.client("secretsmanager")
    password = kwargs["password"]

    if not password and not kwargs["password_secret"]:
        if kwargs["username"] == "elastic":
            password = get_elastic_password("elastic_secret")
        elif kwargs["username"] == "kibana_system":
            password = get_elastic_password("kibana_system_secret")

    if not password:
        if not kwargs["password_secret"]:
            LOG.error("You must specify either --password or --password-secret")
            sys.exit(1)

        response = client.get_secret_value(SecretId=kwargs["password_secret"])
        password = response["SecretString"]

    url = f"{kwargs['es_protocol']}://{kwargs['es_host']}:{kwargs['es_port']}"
    LOG.debug("Connecting to %s", url)
    ctx.obj = {
        "url": url,
        "auth": HTTPBasicAuth(kwargs["username"], password),
        "username": kwargs["username"],
        "password": password,
        "es": Elasticsearch(url, basic_auth=(kwargs["username"], password), request_timeout=kwargs["request_timeout"]),
        "format": kwargs["format"],
    }


for cmd in [cmd_passwd, cmd_cluster_health, cmd_snapshots, cmd_cat, cmd_cluster, cmd_security, cmd_api]:
    # noinspection PyTypeChecker
    ih_elastic.add_command(cmd)



"""
.. topic:: ``ih-ec2 launch-templates``

    A ``ih-ec2 launch-templates`` subcommand.

    See ``ih-ec2 launch-templates`` for more details.
"""

import sys
from logging import getLogger

import click
from botocore.exceptions import ClientError
from tabulate import tabulate

from infrahouse_toolkit.logging import setup_logging

LOG = getLogger()


def list_launch_templates(ec2_client):
    """
    Find and print information about available instance templates in a region.
    """
    response = ec2_client.describe_launch_templates()
    header = ["LaunchTemplateId", "LaunchTemplateName"]
    rows = []
    for launch_template in response["LaunchTemplates"]:
        rows.append([launch_template["LaunchTemplateId"], launch_template["LaunchTemplateName"]])
    print(tabulate(sorted(rows, key=lambda x: x[1]), headers=header, tablefmt="outline"))


@click.command(name="launch-templates")
@click.pass_context
def cmd_launch_templates(ctx):
    """
    Describe AWS launch-templates.
    """
    setup_logging(debug=ctx.obj["debug"])
    ec2_client = ctx.obj["ec2_client"]
    aws_config = ctx.obj["aws_config"]
    try:
        list_launch_templates(ec2_client)
    except ClientError as err:
        LOG.exception(err)
        LOG.info("Try to run ih-ec2 with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)



"""
.. topic:: ``ih-s3-reprepro migrate``

    A ``ih-s3-reprepro migrate`` subcommand.

    See ``ih-s3-reprepro migrate --help`` for more details.
"""

import sys
from logging import getLogger
from os import path as osp

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit.cli.utils import execute, repo_env

LOG = getLogger()


@click.command(name="migrate")
@click.pass_context
def cmd_migrate(ctx: Context):
    """
    Upgrade reprepo database to 5.4.*.
    """
    try:
        with repo_env(
            ctx.parent.params["bucket"],
            ctx.parent.params["role_arn"],
            ctx.parent.params["gpg_key_secret_id"],
            ctx.parent.params["gpg_passphrase_secret_id"],
            region=ctx.parent.params["aws_region"],
        ) as (path, gpg_home):
            LOG.info("Removing existing database %s", osp.join(path, "db/packagenames.db"))
            execute(["rm", "-v", osp.join(path, "db/packagenames.db")])
            LOG.info("Checking the repo in %s", path)
            execute(["reprepro", "-V", "-b", path, "--gnupghome", gpg_home, "check"])
            LOG.info("Exporting the repo in %s", path)
            execute(["reprepro", "-V", "-b", path, "--gnupghome", gpg_home, "export"])

    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



import json
from base64 import b64decode
from os import path as osp

import pytest

from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


@pytest.mark.parametrize(
    "plan_file, result_counts, expected_comment",
    [
        (
            "plan-no-output.stdout",
            (1, 1, 1),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |   🟢 Add |   🟡 Change |   🔴 Destroy |
|----------:|---------:|------------:|-------------:|
|        ✅ |        1 |           1 |            1 |
<details>
<summary>STDOUT</summary>

```
no stdout
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAxLCAiY2hhbmdlIjogMSwgImRlc3Ryb3kiOiAxfX0=
```
</details>""",
        ),
        (
            "plan-0-0-0.stdout",
            (0, 0, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |    Add |    Change |    Destroy |
|----------:|-------:|----------:|-----------:|
|        ✅ |      0 |         0 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform has compared your real infrastructure against your configuration
and found no differences, so no changes are needed.
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAwLCAiY2hhbmdlIjogMCwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
        (
            "plan-2-0-0.stdout",
            (2, 0, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |   🟢 Add |    Change |    Destroy |
|----------:|---------:|----------:|-----------:|
|        ✅ |        2 |         0 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  + create

Terraform will perform the following actions:

  # module.repos["test"].github_repository.repo will be created
  + resource "github_repository" "repo" {
      + allow_auto_merge            = false
      + allow_merge_commit          = true
      + allow_rebase_merge          = true
      + allow_squash_merge          = true
      + archived                    = false
      + default_branch              = (known after apply)
      + delete_branch_on_merge      = false
      + description                 = "Template for a GitHub Control repository"
      + etag                        = (known after apply)
      + full_name                   = (known after apply)
      + git_clone_url               = (known after apply)
      + has_issues                  = true
      + html_url                    = (known after apply)
      + http_clone_url              = (known after apply)
      + id                          = (known after apply)
      + merge_commit_message        = "PR_TITLE"
      + merge_commit_title          = "MERGE_MESSAGE"
      + name                        = "test"
      + node_id                     = (known after apply)
      + private                     = (known after apply)
      + repo_id                     = (known after apply)
      + squash_merge_commit_message = "COMMIT_MESSAGES"
      + squash_merge_commit_title   = "COMMIT_OR_PR_TITLE"
      + ssh_clone_url               = (known after apply)
      + svn_url                     = (known after apply)
      + visibility                  = "public"

      + security_and_analysis {
          + advanced_security {
              + status = (known after apply)
            }

          + secret_scanning {
              + status = (known after apply)
            }

          + secret_scanning_push_protection {
              + status = (known after apply)
            }
        }
    }

  # module.repos["test"].github_team_repository.dev will be created
  + resource "github_team_repository" "dev" {
      + etag       = (known after apply)
      + id         = (known after apply)
      + permission = "push"
      + repository = "test"
      + team_id    = "7332815"
    }

Plan: 2 to add, 0 to change, 0 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAyLCAiY2hhbmdlIjogMCwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
        (
            "plan-2-1-2.stdout",
            (2, 1, 2),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |   🟢 Add |   🟡 Change |   🔴 Destroy |
|----------:|---------:|------------:|-------------:|
|        ✅ |        2 |           1 |            2 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  + create
  ~ update in-place
  - destroy

Terraform will perform the following actions:

  # module.repos["cookiecutter-github-control"].github_repository.repo will be destroyed
  # (because module.repos["cookiecutter-github-control"] is not in configuration)
  - resource "github_repository" "repo" {
      - allow_auto_merge            = false -> null
      - allow_merge_commit          = true -> null
      - allow_rebase_merge          = true -> null
      - allow_squash_merge          = true -> null
      - allow_update_branch         = false -> null
      - archived                    = false -> null
      - default_branch              = "main" -> null
      - delete_branch_on_merge      = false -> null
      - description                 = "Template for a GitHub Control repository" -> null
      - etag                        = "W/\\\"8b4a792bc1474d381caaa63b76668993b4adc42ae76ed53d6886d8562ebb0c67\\\"" -> null
      - full_name                   = "infrahouse/cookiecutter-github-control" -> null
      - git_clone_url               = "git://github.com/infrahouse/cookiecutter-github-control.git" -> null
      - has_discussions             = false -> null
      - has_downloads               = false -> null
      - has_issues                  = true -> null
      - has_projects                = false -> null
      - has_wiki                    = false -> null
      - html_url                    = "https://github.com/infrahouse/cookiecutter-github-control" -> null
      - http_clone_url              = "https://github.com/infrahouse/cookiecutter-github-control.git" -> null
      - id                          = "cookiecutter-github-control" -> null
      - is_template                 = false -> null
      - merge_commit_message        = "PR_TITLE" -> null
      - merge_commit_title          = "MERGE_MESSAGE" -> null
      - name                        = "cookiecutter-github-control" -> null
      - node_id                     = "R_kgDOI528zg" -> null
      - private                     = false -> null
      - repo_id                     = 597540046 -> null
      - squash_merge_commit_message = "COMMIT_MESSAGES" -> null
      - squash_merge_commit_title   = "COMMIT_OR_PR_TITLE" -> null
      - ssh_clone_url               = "git@github.com:infrahouse/cookiecutter-github-control.git" -> null
      - svn_url                     = "https://github.com/infrahouse/cookiecutter-github-control" -> null
      - topics                      = [] -> null
      - visibility                  = "public" -> null
      - vulnerability_alerts        = false -> null

      - security_and_analysis {

          - secret_scanning {
              - status = "disabled" -> null
            }

          - secret_scanning_push_protection {
              - status = "disabled" -> null
            }
        }
    }

  # module.repos["cookiecutter-github-control"].github_team_repository.dev will be destroyed
  # (because module.repos["cookiecutter-github-control"] is not in configuration)
  - resource "github_team_repository" "dev" {
      - etag       = "W/\\\"8043f81b19693f6c1a72d21bb8dc03859c98bf78bfbe79782bfe13fa813992ca\\\"" -> null
      - id         = "7332815:cookiecutter-github-control" -> null
      - permission = "push" -> null
      - repository = "cookiecutter-github-control" -> null
      - team_id    = "7332815" -> null
    }

  # module.repos["infrahouse-toolkit"].github_repository.repo will be updated in-place
  ~ resource "github_repository" "repo" {
      ~ description                 = "InfraHouse Toolkit" -> "InfraHouse Toolkit1"
        id                          = "infrahouse-toolkit"
        name                        = "infrahouse-toolkit"
        # (31 unchanged attributes hidden)

        # (1 unchanged block hidden)
    }

  # module.repos["test"].github_repository.repo will be created
  + resource "github_repository" "repo" {
      + allow_auto_merge            = false
      + allow_merge_commit          = true
      + allow_rebase_merge          = true
      + allow_squash_merge          = true
      + archived                    = false
      + default_branch              = (known after apply)
      + delete_branch_on_merge      = false
      + description                 = "Template for a GitHub Control repository"
      + etag                        = (known after apply)
      + full_name                   = (known after apply)
      + git_clone_url               = (known after apply)
      + has_issues                  = true
      + html_url                    = (known after apply)
      + http_clone_url              = (known after apply)
      + id                          = (known after apply)
      + merge_commit_message        = "PR_TITLE"
      + merge_commit_title          = "MERGE_MESSAGE"
      + name                        = "test"
      + node_id                     = (known after apply)
      + private                     = (known after apply)
      + repo_id                     = (known after apply)
      + squash_merge_commit_message = "COMMIT_MESSAGES"
      + squash_merge_commit_title   = "COMMIT_OR_PR_TITLE"
      + ssh_clone_url               = (known after apply)
      + svn_url                     = (known after apply)
      + visibility                  = "public"

      + security_and_analysis {
          + advanced_security {
              + status = (known after apply)
            }

          + secret_scanning {
              + status = (known after apply)
            }

          + secret_scanning_push_protection {
              + status = (known after apply)
            }
        }
    }

  # module.repos["test"].github_team_repository.dev will be created
  + resource "github_team_repository" "dev" {
      + etag       = (known after apply)
      + id         = (known after apply)
      + permission = "push"
      + repository = "test"
      + team_id    = "7332815"
    }

Plan: 2 to add, 1 to change, 2 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAyLCAiY2hhbmdlIjogMSwgImRlc3Ryb3kiOiAyfX0=
```
</details>""",
        ),
        (
            "plan-0-2-0-a.stdout",
            (0, 2, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |    Add |   🟡 Change |    Destroy |
|----------:|-------:|------------:|-----------:|
|        ✅ |      0 |           2 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  ~ update in-place

Terraform will perform the following actions:

  # module.website.aws_autoscaling_group.website will be updated in-place
  ~ resource "aws_autoscaling_group" "website" {
        id                               = "web20231125205239428700000003"
        name                             = "web20231125205239428700000003"
        # (32 unchanged attributes hidden)

      ~ launch_template {
            id      = "lt-042ea5dd55b0fff3b"
            name    = "web20231125205151213200000001"
          ~ version = "8" -> (known after apply)
        }

        # (9 unchanged blocks hidden)
    }

  # module.website.aws_launch_template.website will be updated in-place
  ~ resource "aws_launch_template" "website" {
        id                                   = "lt-042ea5dd55b0fff3b"
      ~ latest_version                       = 8 -> (known after apply)
        name                                 = "web20231125205151213200000001"
        tags                                 = {}
      ~ user_data                            = " " -> (known after apply)
userdata changes:
--- before
+++ after
@@ -1,109 +1 @@
-Content-Type: multipart/mixed; boundary="MIMEBOUNDARY"
-MIME-Version: 1.0
-
---MIMEBOUNDARY
-Content-Transfer-Encoding: 7bit
-Content-Type: text/cloud-config
-Mime-Version: 1.0
-
-#cloud-config
-"apt":
-  "sources":
-    "infrahouse":
-      "key": |
-        -----BEGIN PGP PUBLIC KEY BLOCK-----
-
-        mQINBGS7FtkBEADF+YmDg6Qvsv5V6ZQw+QhxdTvjMXeLC9T6RFRVD947qKztNNQn
-        +A/yxprLAmW25Uu/11oCszNVIIaR3OY8MtyilbCtzUkBzvWdlX3HvTw91Lp8bxUL
-        EXOM9nscqOrOvfBjanplKDEKPMY6vIZ3kp1rUsCNGLBL0x2KZgU80ulRsMM8qnBN
-        kpTiAeXKOEVtRzlWlvOFLJ7pYDvaZ0UgHSVuE+ZkldqOdLLRpKkdIvJPckXqL5xT
-        UvWTW1enFAHd5+NQxMnCa3ikjQV3DTlEjsgLLQbbysXjXPr6PIPs29cTcfoq431O
-        tR4DrZB4vMEOA88TmKBkEiMG+iB9cjz5wBIuMu95ZmoJxdvjPR72FYKgrCXNtNhd
-        VerERvpLThdaRmZgi7ly+DvkY7LqyJPvdHcqxedgFnPKR4T+gSQmbUPd9qqdJR0Q
-        Kl/PmtuS/w9gy2Z4ay0QIKrNBzQxhA8LDZpp1rufCJ5+ZMef90vHpwxM+OaPryNl
-        EdDW7YQ8fuB5nPqjwUApmlYYLJYld7vJ79EIx/sjgh6G+IFim7xR7WiQF/nJevfq
-        vzYO7or/GICb6eZwvPdty79HmVePqETlDU00oyopWyTsYySYDSHn/4doTux8v+F/
-        YeybvgqbriH7iSF8HxwEFeDjrMtUTHsXC35DHFeMChJoxsyKv48J+bIGFwARAQAB
-        tC1JbmZyYUhvdXNlIFBhY2thZ2VyIDxwYWNrYWdlckBpbmZyYWhvdXNlLmNvbT6J
-        AlQEEwEKAD4WIQSiNh5kY3w8W0gAF8XguJaoxfcD7wUCZLsW2QIbAwUJA8JnAAUL
-        CQgHAgYVCgkICwIEFgIDAQIeAQIXgAAKCRDguJaoxfcD7wb7D/9enbtO1DEQn1Pt
-        hgktH+IrbBdYRHk71il2hbnmAboLkk1ak/aaQY/XQDJFhrA7jMkPLOCReotHiQmv
-        1UfqDcM2eQ4/x0N9F2DLUSkuww117vttUoBxirB72pO0T/ld9o39SHMC/Dfk6y6h
-        tTGqxLgh8j3DSyO5EErJLa/3BvMqr4iqb3PTwNeDREjavQJhXSvEFjyACoo9d7rs
-        7lhCbbSPjgTcG/PCfzLr37QyzXiTR/TwjiwVdThdCS/CKnsoM8ngHxjXYpK7M6wg
-        tzA8JkHZr5MpkrPGrgQevhJBLkHR/jcsaaBZN7nw8aeHcwwM4/ChHIyZdQ91059p
-        OMaYmWmjPQfq1rOnAAvA1xWw+TZNzHgWIq4vusx7OBvrFsVeZNsqd7Oa4prtVWg0
-        91cKZJo/Ov6opM/wDWN2yBnkmBCwaXpoH5NDQ22bhnGTJeIWdei8USGpvBkTxBI2
-        X32pyOmZDn81L2zM2ioNvgXIb5u5tjb9NOiAt6oW0G0x00dgr3r6pu8qmMwkBVBC
-        vWwEmaxQnkThnbAd57Vbat4//uDcHFpDaBPpFWeCL2dfDSrS9iiK6K1VkvGLoXh8
-        MpJqjoddpfNh4uyfzKyNE1drQ29mcwfIlzXoPrb09W6ZHpm+tqd8Mk707zwsdFGR
-        qRsxmtPn9J1reYZiVuE5G26f1aKTR7kCDQRkuxbZARAA4+9kn4smBXZcM0vFq9R8
-        24XKZtTaVmVoVJMRQM44qWR+s/S0MzvbO5u0cSVThY65YzMLs920UY6uGpoS2ZYw
-        IMipkaB+2F6B8R7V7nl/SgaXGa2MexkC7wt6blFbaw7pbvI+iI9wS8BKHgMDzd0F
-        TjSq9vdkVkf7yI3hEwllX7zT6ILOvknNGbos0dFI92cBWQs+iXP50xDVit3+9mCD
-        ruWVYk73mvSEzFcuXCybKJ5U/nzSQ70JoQmSPc50RQjuHGdzl/YzGFaFn5zIZ5pP
-        k5S+KTYekzciRkHzbea5Pj+A6KZilKj5ykARFzw+CPxqsuejDki7A+a3m3WGq+mm
-        AklQ4YBcolzNOpgWclISuE3d5kgtaJACIgte87F6QBBsk8oDBFQCnZmoOirlSlbf
-        JuD6fVCJu2zY7ThN1O0oWTwAaNAvM55CLTv6s3FmtR92lnlQDFuJ+fj863c6qx14
-        FuO4mM7euS08wGgVGm91wA7bYl4ahxC6S2BGB5FvkZfzBBwjsfmUWHAu7VRZJxkX
-        Jp696OTdR3RfI7IZ6OW6tGELijswWbbZXb5XRWEW99vWO7NIK6sOvhMDNp9bSMub
-        XP3PbPxgyT0bmss0ywNYwucnfcmDXqPVu3Gdt53at1+hK7mlnQ3uUs5eVEOIHnJm
-        f1trDggSTE7fnkx6Czl+iMkAEQEAAYkCPAQYAQoAJhYhBKI2HmRjfDxbSAAXxeC4
-        lqjF9wPvBQJkuxbZAhsMBQkDwmcAAAoJEOC4lqjF9wPv8SQP/jsQfJCUHrjpX4fP
-        0NtVyqlCrq1OmLP7BLrXrCxoAFRb64YJUdWk5P35k9wj2BQjpjcNj2ik80Foo02E
-        yc+PzW/4tzVvdxRrEzsOFm+mMR+llUHAKzBa6QBxHT0ofL7p5XYTuMNnYo+PjVMm
-        jP3TEJGYXYmuhGfaaJZpjLmJFqp59x//dsMHeLqUChwb6Cg6DukAiheAs3P+rXo3
-        tBnA4JCCiXR7//0WezRoSPdalNem4vs8PT87snHbDliFMVf8w+fksInqvotiRDI4
-        QBrBfs3NdeZW/tmupOaZJRzHDvvKbg//O6ip1aPvILWvTDM8KcQCQESaWkCbI1uc
-        rizAMblTB2gyw9yojCpVCMBXdizAqqyl5ZVrw4+X/i09IsNt8hA9/5RY7v3c7Z2C
-        Xy2xKWmtLI9nlOSSwUxB566I29lVKBLUmXmQfnRIIHCiJDXYFDP/6Lfv9+mhEWU8
-        qed3xhpeNSAIzA+MTljtiRmorxpddiXbtCafcPl/J3X0B07fccJZwZsCYCptLy1c
-        yvVaN+EBmjoCp3jt6wLe4xclqU3ACwz/8TDDmtJ7UUdl7MdkipQZ278L2kpl6o61
-        MF2fXD7fZMeZfEw6lu0BF5mZ0/ch3HxXeZoXif79g25vZmSdW97pceewPlvlETFQ
-        55pMorQyUTEQCQ3ECzHYSnL/TmW7
-        =+0Rc
-        -----END PGP PUBLIC KEY BLOCK-----
-      "source": "deb [signed-by=$KEY_FILE] https://release-jammy.infrahouse.com/ $RELEASE
-        main"
-"package_update": true
-"packages":
-- "make"
-- "gcc"
-- "puppet-code"
-- "infrahouse-toolkit"
-- "infrahouse-puppet-data"
-"puppet":
-  "collection": "puppet8"
-  "install": true
-  "install_type": "aio"
-  "package_name": "puppet-agent"
-  "start_service": false
-"runcmd":
-- "/opt/puppetlabs/puppet/bin/gem install json"
-- "/opt/puppetlabs/puppet/bin/gem install aws-sdk-core"
-- "/opt/puppetlabs/puppet/bin/gem install aws-sdk-secretsmanager"
-- "ih-puppet  --environment production --environmentpath {root_directory}/environments
-  --root-directory /opt/puppet-code --hiera-config /opt/infrahouse-puppet-data/environments/production/hiera.yaml
-  --module-path {root_directory}/modules apply /opt/puppet-code/environments/production/manifests/site.pp"
-"write_files":
-- "content": "export AWS_DEFAULT_REGION=us-west-1"
-  "path": "/etc/profile.d/aws.sh"
-  "permissions": "0644"
-- "content": |-
-    [default]
-    region=us-west-1
-  "path": "/root/.aws/config"
-  "permissions": "0600"
-- "content": |
-    "puppet_environment": "production"
-    "puppet_role": "webserver"
-  "path": "/etc/puppetlabs/facter/facts.d/puppet.yaml"
-  "permissions": "0644"
-- "content": "{\\"ih-puppet\\":{\\"debug\\":false,\\"environmentpath\\":\\"{root_directory}/environments\\",\\"hiera-config\\":\\"/opt/infrahouse-puppet-data/environments/production/hiera.yaml\\",\\"manifest\\":\\"/opt/puppet-code/environments/production/manifests/site.pp\\",\\"module-path\\":\\"{root_directory}/modules\\",\\"root-directory\\":\\"/opt/puppet-code\\"}}"
-  "path": "/etc/puppetlabs/facter/facts.d/ih-puppet.json"
-  "permissions": "0644"
-- "content": "{}"
-  "path": "/etc/puppetlabs/facter/facts.d/custom.json"
-  "permissions": "0644"
-
---MIMEBOUNDARY--
+(known after apply)
EOF userdata changes.
        # (16 unchanged attributes hidden)

      + metadata_options {
          + http_tokens = "required"
        }

        # (4 unchanged blocks hidden)
    }

Plan: 0 to add, 2 to change, 0 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAwLCAiY2hhbmdlIjogMiwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
        (
            "plan-0-2-0.stdout",
            (0, 2, 0),
            """
# State **`s3://foo_backet/path/to/tf.state`**
## Affected resources counts

|   Success |    Add |   🟡 Change |    Destroy |
|----------:|-------:|------------:|-----------:|
|        ✅ |      0 |           2 |          0 |
<details>
<summary>STDOUT</summary>

```
Terraform used the selected providers to generate the following execution
plan. Resource actions are indicated with the following symbols:
  ~ update in-place

Terraform will perform the following actions:

  # module.website.aws_autoscaling_group.website will be updated in-place
  ~ resource "aws_autoscaling_group" "website" {
        id                               = "web20231125205239428700000003"
        name                             = "web20231125205239428700000003"
        # (32 unchanged attributes hidden)

      ~ launch_template {
            id      = "lt-042ea5dd55b0fff3b"
            name    = "web20231125205151213200000001"
          ~ version = "8" -> (known after apply)
        }

        # (9 unchanged blocks hidden)
    }

  # module.website.aws_launch_template.website will be updated in-place
  ~ resource "aws_launch_template" "website" {
        id                                   = "lt-042ea5dd55b0fff3b"
      ~ latest_version                       = 8 -> (known after apply)
        name                                 = "web20231125205151213200000001"
        tags                                 = {}
      ~ user_data                            = " " -> " "
userdata changes:
--- before
+++ after
@@ -99,7 +99,7 @@
     "puppet_role": "webserver"
   "path": "/etc/puppetlabs/facter/facts.d/puppet.yaml"
   "permissions": "0644"
-- "content": "{\\"ih-puppet\\":{\\"debug\\":false,\\"environmentpath\\":\\"{root_directory}/environments\\",\\"hiera-config\\":\\"/opt/infrahouse-puppet-data/environments/production/hiera.yaml\\",\\"manifest\\":\\"/opt/puppet-code/environments/production/manifests/site.pp\\",\\"module-path\\":\\"{root_directory}/modules\\",\\"root-directory\\":\\"/opt/puppet-code\\"}}"
+- "content": "{\\"ih-puppet\\":{\\"cancel_instance_refresh_on_error\\":false,\\"debug\\":false,\\"environmentpath\\":\\"{root_directory}/environments\\",\\"hiera-config\\":\\"/opt/infrahouse-puppet-data/environments/production/hiera.yaml\\",\\"manifest\\":\\"/opt/puppet-code/environments/production/manifests/site.pp\\",\\"module-path\\":\\"{root_directory}/modules\\",\\"root-directory\\":\\"/opt/puppet-code\\"}}"
   "path": "/etc/puppetlabs/facter/facts.d/ih-puppet.json"
   "permissions": "0644"
 - "content": "{}"
EOF userdata changes.
        # (16 unchanged attributes hidden)

      + metadata_options {
          + http_tokens = "required"
        }

        # (4 unchanged blocks hidden)
    }

Plan: 0 to add, 2 to change, 0 to destroy.

─────────────────────────────────────────────────────────────────────────────

Saved the plan to: tf.plan

To perform exactly these actions, run the following command to apply:
    terraform apply "tf.plan"
```
</details>
<details><summary><i>metadata</i></summary>

```
eyJzMzovL2Zvb19iYWNrZXQvcGF0aC90by90Zi5zdGF0ZSI6IHsic3VjY2VzcyI6IHRydWUsICJhZGQiOiAwLCAiY2hhbmdlIjogMiwgImRlc3Ryb3kiOiAwfX0=
```
</details>""",
        ),
    ],
)
def test_comment(plan_file, result_counts, expected_comment):
    with open(osp.join("infrahouse_toolkit/terraform/tests/plans", plan_file)) as fp:
        status = TFStatus(
            TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(*result_counts), RunOutput(fp.read(), None)
        )
        assert isinstance(status.comment, str)
        print("\nActual comment:")
        print(status.comment)
        print("EOF Actual comment.")
        assert status.comment == expected_comment


def test_comment_none():
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        RunResult(None, None, None),
        RunOutput("no stdout", "no stderr"),
        affected_resources=RunResult(None, None, None),
    )
    assert isinstance(status.comment, str)
    print(status.comment)
